from configparser import ConfigParser

from simses.commons.data.data_handler import DataHandler
from simses.commons.state.technology.redox_flow import RedoxFlowState
from simses.system.auxiliary.auxiliary import Auxiliary
from simses.system.auxiliary.pump.abstract_pump import Pump
from simses.technology.redox_flow.degradation.capacity_degradation import CapacityDegradationModel
from simses.technology.redox_flow.electrochemical.control.redox_control_system import RedoxControlSystem
from simses.technology.redox_flow.factory import RedoxFlowFactory
from simses.technology.redox_flow.pump_algorithm.abstract_pump_algorithm import PumpAlgorithm
from simses.technology.redox_flow.stack.abstract_stack import StackModule
from simses.technology.redox_flow.stack.electrolyte.abstract_electrolyte import ElectrolyteSystem
from simses.technology.storage import StorageTechnology


class RedoxFlow(StorageTechnology):
    """The RedoxFlow class updates the electrochemical model that includes the redox control system, the
    degradation model and the pump algorithm."""

    __MAX_LOOPS = 10
    __ACCURACY = 1e-9

    def __init__(self, stack_type, power, voltage, capacity, pump_algorithm, temperature, data_export: DataHandler,
                 storage_id: int, system_id: int, config: ConfigParser):
        super().__init__()
        self.__factory: RedoxFlowFactory = RedoxFlowFactory(config)
        self.__max_power = power
        self.__capacity = capacity
        self.__electrolyte_system: ElectrolyteSystem = self.__factory.create_electrolyte_system(capacity, stack_type)
        self.__stack_module: StackModule = self.__factory.create_stack_module(stack_type, self.__electrolyte_system,
                                                                              voltage, power)
        self.__capacity_degradation_model: CapacityDegradationModel = self.__factory.create_degradation_model(capacity,
            self.__stack_module)
        pump_algorithm = self.__factory.check_pump_algorithm(pump_algorithm, stack_type)
        self.__pump: Pump = self.__factory.create_pumps(pump_algorithm)
        self.__pump_algorithm: PumpAlgorithm = self.__factory.create_pump_algorithm(self.__pump, self.__stack_module,
                                                                                    self.__electrolyte_system,
                                                                                    pump_algorithm)
        battery_management_system: RedoxControlSystem = self.__factory.create_redox_management_system(
            self.__stack_module, self.__electrolyte_system, self.__pump_algorithm)
        self.__redox_flow_state: RedoxFlowState = self.__factory.create_redox_flow_state_from(storage_id, system_id,
                                                                                              self.__stack_module,
                                                                                              capacity, temperature)
        self.__electrochemical_model = self.__factory.create_electrochemical_model(self.__stack_module,
                                                                                   battery_management_system,
                                                                                   self.__electrolyte_system,
                                                                                   self.__pump_algorithm)
        self.__time = self.__redox_flow_state.time
        self.__data_export: DataHandler = data_export
        self.__data_export.transfer_data(self.__redox_flow_state.to_export())

    def update(self):
        """
        Starts updating the calculation for the electrochemical model of the redox flow battery, which includes the
        battery management system requests.

        Returns
        -------
            None

        """
        rfbs = self.__redox_flow_state
        time = self.__time
        soc_start_time_step = self.__redox_flow_state.soc
        soc_stack_start_time_step = self.__redox_flow_state.soc_stack
        self.get_equilibrium_state_for(time, rfbs, soc_start_time_step, soc_stack_start_time_step)
        self.__pump_algorithm.update(rfbs, soc_start_time_step)
        self.__capacity_degradation_model.update(time, rfbs)
        rfbs.time = time
        self.__data_export.transfer_data(rfbs.to_export())

    def get_equilibrium_state_for(self, time: float, rfbs: RedoxFlowState, soc_start_time_step: float,
                                  soc_stack_start_time_step: float):
        for step in range(self.__MAX_LOOPS):
            old_resistance = rfbs.internal_resistance
            rfbs.soc_stack = soc_stack_start_time_step
            rfbs.soc = soc_start_time_step
            self.__electrochemical_model.update(time, rfbs)
            if (not self.__stack_module.dependent_parameters() or
                    (abs(rfbs.internal_resistance - old_resistance) < self.__ACCURACY)):
                break

    def set(self, time: float, current: float, voltage: float):
        """
        Sets the new simulation time an sets the power (current * voltage) of the RedoxFlowState for the next simulation
        time step.

        Parameters
        ----------
        time : float
            current time of the simulation
        current : float
            target current in A
        voltage : float
            target voltage in V

        Returns
        -------
            None
        """
        self.__time = time
        self.__redox_flow_state.power_in = current * voltage

    def distribute_and_run(self, time: float, current: float, voltage: float):
        self.set(time, current, voltage)
        self.update()

    @property
    def volume(self) -> float:
        """
        Volume of storage technology in m^3.

        Returns
        -------
        float:
            Redox flow volume in m^3.
        """
        """
        The volume of electrolyte and stacks are considered. 
        """
        thickness_end_plate = 0.06  # m
        # Geometric factor that describes by how much the flow frame (length x width) is larger than the electrode area
        flow_frame_dimension_factor = 1.5  # m

        energy_density_electrolyte = (self.__electrolyte_system.get_capacity_density() *
                                      self.__stack_module.get_nominal_voltage_cell())
        volume_electrolyte = self.__capacity * 3600 / energy_density_electrolyte

        volume_stacks = (((self.__stack_module.get_electrode_thickness() * self.__stack_module.get_cell_per_stack()) *
                         2 + 2 * thickness_end_plate) * self.__stack_module.get_specific_cell_area() / 10000 *
                         flow_frame_dimension_factor * self.__stack_module.get_parallel_scale() *
                         self.__stack_module.get_serial_scale())

        volume_system = volume_electrolyte + volume_stacks
        return volume_system

    @property
    def mass(self) -> float:
        """
        Mass of the storage technology in kg.

        Returns
        -------
        float:
            Mass in kg.
        """
        """
        The mass of electrolyte and stacks are considered. For the stack, an equation was established that takes into 
        account the mass of aluminum end plates, copper current collectors, and polypropylene flow frames, which are 
        scaled with the electrode area and the number of cells. 
        """
        c = 3.22E2
        d = 4.65E2

        energy_density_electrolyte_mass = (self.__electrolyte_system.get_capacity_density() *
                                           self.__stack_module.get_nominal_voltage_cell() /
                                           self.__electrolyte_system.get_electrolyte_density())
        mass_electrolyte = self.__capacity * 3600 / energy_density_electrolyte_mass  # kg

        mass_stacks = ((c + d * self.__stack_module.get_electrode_thickness() *
                       self.__stack_module.get_cell_per_stack()) * self.__stack_module.get_specific_cell_area() / 10000
                       * self.__stack_module.get_serial_scale() * self.__stack_module.get_parallel_scale())  # kg

        mass_system = mass_stacks + mass_electrolyte
        return mass_system

    @property
    def surface_area(self) -> float:
        """
        Surface area of the storage technology that can be impacted by confection m^2.

        Returns
        -------
        float:
            Surface area of the storage technology in m^2.
        """
        """
        The open surface area of the tanks and stacks are considered.
        Assumption: The tank volume is box-shaped and in the bottom of the housing. The bottom is facing the ground, 
        but the other walls are open to air flow. The stacks are mounted in racks and all sides are open to air flow.
        """
        thickness_end_plate = 0.06
        flow_frame_dimension_factor = 1.5

        energy_density_electrolyte = (self.__electrolyte_system.get_capacity_density() *
                                      self.__stack_module.get_nominal_voltage_cell())
        volume_electrolyte = self.__capacity * 3600 / energy_density_electrolyte
        width_tanks = volume_electrolyte ** (1 / 3)
        surface_tanks = 5 * width_tanks ** 2

        length_stack = ((self.__stack_module.get_electrode_thickness() * self.__stack_module.get_cell_per_stack()) *
                        2 + 2 * thickness_end_plate)
        width_stack = (self.__stack_module.get_specific_cell_area() / 10000 * flow_frame_dimension_factor * 2) ** (1/2)

        surface_stacks = (length_stack * width_stack * 3 + width_stack ** 2 * self.__stack_module.get_serial_scale() *
                          self.__stack_module.get_parallel_scale())

        surface_system = surface_tanks + surface_stacks
        return surface_system

    @property
    def specific_heat(self) -> float:
        """
        Specific heat of storage technology in J/(kgK).

        Returns
        -------
        float:
            Specific heat capacity in J/[kgK).
        """
        """
        The specific heat is a mean value for the electrolyte and stacks. 
        For the specific heat of the stack, an equation was established that takes into account the aluminum end plates, 
        copper current collectors, and polypropylene flow frames.
        """
        a = 2.84E5
        b = 7.91E5
        c = 3.22E2
        d = 4.65E2

        energy_density_electrolyte_mass = (self.__electrolyte_system.get_capacity_density() *
                                           self.__stack_module.get_nominal_voltage_cell() /
                                           self.__electrolyte_system.get_electrolyte_density())

        # Hoffmann, Lisa. Physical properties of a VRFB-electrolyte and their impact on the cell-performance.
        specific_heat_electrolyte = 2960  # J/(kgK)

        specific_heat_stack_mean = ((a + b * self.__stack_module.get_electrode_thickness() *
                                    self.__stack_module.get_cell_per_stack()) / (c * d *
                                    self.__stack_module.get_electrode_thickness() *
                                    self.__stack_module.get_cell_per_stack()))  # J/(kgK)

        mass_electrolyte = self.__capacity * 3600 / energy_density_electrolyte_mass  # kg
        mass_stacks = ((c + d * self.__stack_module.get_electrode_thickness() *
                       self.__stack_module.get_cell_per_stack()) * self.__stack_module.get_specific_cell_area() / 10000
                       * self.__stack_module.get_serial_scale() * self.__stack_module.get_parallel_scale())  # kg

        specific_heat_system = ((specific_heat_electrolyte * mass_electrolyte + specific_heat_stack_mean + mass_stacks)
                                / (mass_stacks + mass_electrolyte))

        return specific_heat_system

    @property
    def convection_coefficient(self) -> float:
        """
        Convective heat transfer coefficient of a redox flow battery.

        A value of 5 W/m^2/K is assumed for the convective heat transfer coefficient, which is in the same order of
        magnitude as the reported values from Tang et al. (2012) and Yan et al. (2016).
        If the thermal model is to be used for a particular redox flow battery geometry, the value must be adjusted with
        respect to the system specifications.
        Tang, Ao, et al. "Thermal modelling and simulation of the all-vanadium redox flow battery." Journal of Power
        Sources 203 (2012): 165-176.
        Yan, Yitao, et al. "Modelling and simulation of thermal behaviour of vanadium redox flow battery." Journal of
        Power Sources 322 (2016): 116-128.

        Returns
        -------
        float:
            Convective heat transfer coefficient in W/(m^2*K).
        """
        return 5

    def wait(self):
        pass

    def get_auxiliaries(self) -> [Auxiliary]:
        return [self.__pump]

    @property
    def state(self) -> RedoxFlowState:
        return self.__redox_flow_state

    def get_system_parameters(self) -> dict:
        return dict()

    def close(self):
        """Closing all resources in redox_flow_system"""
        self.__factory.close()



from simses.commons.cycle_detection.cycle_detector import CycleDetector
from simses.commons.log import Logger
from simses.commons.state.technology.lithium_ion import LithiumIonState
from simses.technology.lithium_ion.cell.type import CellType
from simses.technology.lithium_ion.degradation.cyclic.cyclic_degradation import \
    CyclicDegradationModel


class PanasonicNCACyclicDegradationModel(CyclicDegradationModel):

    def __init__(self, cell_type: CellType, cycle_detector: CycleDetector):
        super().__init__(cell_type, cycle_detector)
        self.__log: Logger = Logger(type(self).__name__)
        self.__capacity_loss = 0
        self.__capacity_loss_cyclic = cell_type.get_cyclic_capacity_loss_start()
        self.__fec = 0
        self.__rinc_cyclic = 0
        self.__resistance_increase = 0

        #  Aging Model based on:
        #  Geng, Yuhui (2019): Aging Model of Lithium-ion Battery. Master Thesis. Technical University Munich, EES.

        self.__A_QLOSS = 0.007043  # constant
        self.__B_QLOSS = -0.008682  # constant
        self.__C_QLOSS = 0.005818  # constant
        self.__D_QLOSS = 0.001892  # constant
        self.__E_QLOSS = 0
        self.__F_QLOSS = -0.001609

        self.__A_RINC = -0.09964  # constant
        self.__B_RINC = 0.03746  # constant
        self.__C_RINC = 0.01577  # constant
        self.__D_RINC = -0.001825  # constant
        self.__E_RINC = 0.001612  # constant
        self.__F_RINC = -0.006829  # constant

        self.__voltage_last_step = 3.6

    def calculate_degradation(self, battery_state: LithiumIonState) -> None:
        crate: float = self._cycle_detector.get_crate() * 3600 # in 1 / s -> *3600 -> in 1/h
        delta_fec: float = self._cycle_detector.get_delta_full_equivalent_cycle() # Delta EFC
        self.__fec = self._cycle_detector.get_full_equivalent_cycle() # Updating FEC

        qloss: float = self.__capacity_loss_cyclic # in p.u.

        # Open circuit voltage is used for cyclic degradation model
        voltage: float = self._cell.get_open_circuit_voltage(battery_state) / self._cell.get_serial_scale()
        v_ch = max(self.__voltage_last_step, voltage)  # charge voltage -> defined as higher one of both
        v_disch = min(self.__voltage_last_step, voltage)  # charge voltage -> defined as lower one of both

        if crate < 0.3571:
            virtual_fec: float = (qloss / (self.__A_QLOSS + self.__B_QLOSS * v_ch + self.__C_QLOSS * v_disch +
                                           self.__D_QLOSS * v_ch ** 2 + self.__F_QLOSS * v_ch * v_disch)) ** (4 / 3)
            rel_capacity_loss = lambda fec: (self.__A_QLOSS + self.__B_QLOSS * v_ch + self.__C_QLOSS * v_disch +
                                             self.__D_QLOSS * v_ch ** 2 + self.__F_QLOSS * v_ch * v_disch) * fec ** 0.75

            fec = virtual_fec + delta_fec
            capacity_loss = rel_capacity_loss(fec) - self.__capacity_loss_cyclic
        else:
            virtual_fec: float = (qloss / ((1 + (crate - 0.3571) / (0.3571*3)) *
                                           (self.__A_QLOSS + self.__B_QLOSS * v_ch + self.__C_QLOSS * v_disch +
                                            self.__D_QLOSS * v_ch ** 2 + self.__F_QLOSS * v_ch * v_disch))) ** (4 / 3)
            rel_capacity_loss = lambda fec: (1 + (crate - 0.3571) / (3*0.3571)) * \
                                            (self.__A_QLOSS + self.__B_QLOSS * v_ch + self.__C_QLOSS * v_disch +
                                             self.__D_QLOSS * v_ch ** 2 + self.__F_QLOSS * v_ch * v_disch) * fec ** 0.75

            fec = virtual_fec + delta_fec
            capacity_loss = rel_capacity_loss(fec) - self.__capacity_loss_cyclic

        self.__capacity_loss_cyclic += capacity_loss  # p.u.
        self.__capacity_loss = capacity_loss * self._cell.get_nominal_capacity()  # in Ah

        self.__voltage_last_step = voltage # save voltage for next calculation

    def calculate_resistance_increase(self, battery_state: LithiumIonState) -> None:
        delta_fec: float = self._cycle_detector.get_delta_full_equivalent_cycle()  # Delta EFC
        rinc: float = self.__rinc_cyclic # in p.u.

        # Open circuit voltage is used for cyclic degradation model
        voltage: float = self._cell.get_open_circuit_voltage(battery_state) / self._cell.get_serial_scale()
        v_ch = max(self.__voltage_last_step, voltage)  # charge voltage -> defined as higher one of both
        v_disch = min(self.__voltage_last_step, voltage)  # charge voltage -> defined as lower one of both

        virtual_fec = rinc / (
                self.__A_RINC + self.__B_RINC * v_ch + self.__C_RINC * v_disch + self.__D_RINC * v_ch ** 2 +
                self.__E_RINC * v_ch * v_disch + self.__F_RINC * v_disch ** 2)

        rel_resistance_increase = lambda fec: (self.__A_RINC + self.__B_RINC * v_ch + self.__C_RINC * v_disch +
                                               self.__D_RINC * v_ch ** 2 + self.__E_RINC * v_ch * v_disch +
                                               self.__F_RINC * v_disch ** 2) * fec

        resistance_increase = rel_resistance_increase(virtual_fec + delta_fec) - self.__rinc_cyclic

        if resistance_increase < 0: # No data for resistance increase at low SOC. For further information see
            # Geng, Yuhui (2019): Aging Model of Lithium-ion Battery. Master Thesis. Technical University Munich, EES.
            self.__log.warn('NCA cyclic resistance increase model is invalid when: 1. both charge and discharge voltage'
                            ' are close to lower operation voltage limit; 2. by high DoD (> 50 %). In this case'
                            ' the resistance increase is set to 0 for current timestep')
        resistance_increase = max(resistance_increase, 0)

        self.__rinc_cyclic += resistance_increase

        self.__resistance_increase = resistance_increase  # pu

    def get_degradation(self) -> float:
        capacity_loss = self.__capacity_loss
        self.__capacity_loss = 0  # Set value to 0, because cyclic losses are not calculated in each step
        return capacity_loss

    def get_resistance_increase(self) -> float:
        resistance_increase = self.__resistance_increase
        self.__resistance_increase = 0  # Set value to 0, because cyclic losses are not calculated in each step
        return resistance_increase

    def reset(self, lithium_ion_state: LithiumIonState) -> None:
        self.__capacity_loss = 0
        self.__capacity_loss_cyclic = 0
        self.__fec = 0
        self.__rinc_cyclic = 0
        self.__resistance_increase = 0

    def close(self) -> None:
        self.__log.close()



import datetime
from abc import abstractmethod, ABC
import plotly
from simses.analysis.evaluation.plotting.axis import Axis

class Plotting(ABC):

    class Color:
        BLUE = "#0065BD"
        DARK_SKY_BLUE = '#8CBED6'
        ROYAL_BLUE = '#4169E1'
        YELLOW = '#FFDC00'
        GREEN = '#A2AD00'
        BRIGHT_GREEN = '#66FF00'
        RED = '#FF4136'
        CYAN = '#00FFFF'
        MAGENTA = '#F012BE'
        BLACK = '#000000'
        WHITE = '#FFFFFF'
        BROWN = '#964B00'
        PURPLE = '#800080'
        VIOLET = '#EE82EE'
        GREY = '#D3D3D3'
        BRIGHT_BLUE = '#0096FF'

        SOC_BLUE = "#0065BD"
        STACK_PURPLE = "#0E103D"
        CATHODE_PINK = "#C585B3"
        ANODE_GREEN = "#4A5240"
        TEMPERATURE_RED = "#AF1B3F"
        SOH_GREEN = "#A2AD00"
        HEAT_ORANGE = "#FFBC42"
        AC_POWER_BLUE = "#0496FF"
        DC_POWER_GREEN = "#06D6A0"
        RESISTANCE_BLACK = "#000000"
        CURRENT_CYAN = '#98C6EA'
        VOLTAGE_GREEN = "#A2AD00"
        POWER_YELLOW = "#FFDC00"
    """
    Plotting is an abstract class providing the interfaces for different plotting engines.
    """

    def __init__(self):
        self.__figures: list = list()

    @abstractmethod
    def lines(self, xaxis: Axis, yaxis: [Axis], secondary: list = []):
        """
        Creates a figure object by adding traces from the passed axes.

        Parameters:
            yaxis: List of y-axes.
            secondary: List of secondary axes.
        """
        pass

    def get_figures(self) -> list:
        """
        Returns the list of figures saved in the instance of the plotting class instance.
        """
        return self.__figures

    @abstractmethod
    def histogram(self, xaxis: Axis, yaxis: [Axis]):
        """
        Creates a histogram object by adding traces from the passed axes.

        Parameters:
            xaxis: x-axis.
            yaxis: List of y-axes.
        """
        pass

    @abstractmethod
    def bar(self, yaxis: [Axis], bars: int):
        """
        Creates a bar plot by adding traces from the passed axes.

        Parameters:
            yaxis: List of y-axes.
            bars: Number of bars per figure
        """
        pass

    @abstractmethod
    def sankey_diagram(self, node_links: dict):
        """
        Creates a sankey diagram from the passed nodes and links
        :param node_links: dict containing source nodes, target nodes, and values for links
        """
        pass

    def sunburst_diagram(self, categories: dict):
        """
        Creates a sunburst diagram for the energetic losses
        :param categories: dict containing labels, parents, and values
        :return:
        """
        pass

    @abstractmethod
    def subplots(self, xaxis: Axis, yaxis: [Axis]):
        """
        Creates a figure object consisting of subplots from passed axes.

        Parameters:
            xaxis: x-Axis.
            yaxis: List of y-axes.
        """
        pass

    def alphanumerize(self, string) -> str:
        """
        Returns a valid alphanumeric string that can be used for a filename.

        Parameters:
            string: String to be processed.
        """
        return ''.join(e for e in string if e.isalnum())

    @staticmethod
    def convert_to_html(figure) -> str:
        """
        Returns a string that can be embedded in an html from a passed figure object.

        Parameters:
            figure: Figure to be converted to a html-readable string.
        """
        if isinstance(figure, plotly.graph_objs.Figure):
            return figure.to_html(auto_play=False,
                                  include_plotlyjs=True,
                                  include_mathjax=False,
                                  # post_script=plot_id,
                                  full_html=False,
                                  # default_height=()),
                                  validate=True
                                  )

    @staticmethod
    def format_time(time_data):
        """
        Converts list of timestamps into a list of datetimes.

        Parameters:
            time_data: List of timestamps.
        """
        time = list()
        for tstmp in time_data:
            time.append(datetime.datetime.fromtimestamp(tstmp, tz=datetime.timezone.utc))
        return time



import math
from datetime import datetime

from simses.system.auxiliary.auxiliary import Auxiliary
from simses.system.dc_coupling.generation.dc_generation import DcGeneration


class PvDcGeneration(DcGeneration):

    """
    PvDcGeneration provides a basic algorithm imitating a pv plant with the same power each day.
    """

    def __init__(self, peak_power: float):
        super().__init__()
        self.__power = 0.0
        self.__peak_power: float = peak_power
        self.__peak_time: float = 12.0  # hour of day
        self.__time_variance: float = 1.0  # hour

    def get_power(self) -> float:
        return self.__power

    def calculate_power(self, time: float) -> None:
        tstmp = datetime.utcfromtimestamp(time)
        hour = tstmp.hour + tstmp.minute / 60. + tstmp.second / 3600.
        if 6 < hour < 18:
            self.__power = self.__peak_power * self.__daily_power_distribution_of(hour)
        else:
            self.__power = 0.0

    def __daily_power_distribution_of(self, hour: float) -> float:
        return math.exp(-.5 * ((hour - self.__peak_time) / self.__time_variance)**2) \
               / (self.__time_variance * math.sqrt(2*math.pi))

    def get_auxiliaries(self) -> [Auxiliary]:
        return list()

    def close(self):
        pass



from configparser import ConfigParser

import pytest

from simses.commons.config.data.redox_flow import RedoxFlowDataConfig
from simses.commons.config.simulation.redox_flow import RedoxFlowConfig
from simses.commons.state.technology.redox_flow import RedoxFlowState
from simses.commons.utils.utilities import all_non_abstract_subclasses_of
from simses.technology.redox_flow.stack.abstract_stack import StackModule
from simses.technology.redox_flow.stack.cell_data_stack_5500w import CellDataStack5500W
from simses.technology.redox_flow.stack.dummy_stack_3000w import DummyStack3000W
from simses.technology.redox_flow.stack.dummy_stack_5500W import DummyStack5500W
from simses.technology.redox_flow.stack.electrolyte.abstract_electrolyte import ElectrolyteSystem
from simses.technology.redox_flow.stack.electrolyte.vanadium import VanadiumSystem
from simses.technology.redox_flow.stack.high_performance_stack_5700W import HighPerformanceStack5700W
from simses.technology.redox_flow.stack.industrial_stack_1500w import IndustrialStack1500W
from simses.technology.redox_flow.stack.industrial_stack_9000w import IndustrialStack9000W


class TestClassStackModule:

    redox_flow_state: RedoxFlowState = RedoxFlowState(0, 0)
    redox_flow_state.soc = 0.5

    def make_stack_module(self, stack_module_typ_subclass):
        voltage = 20
        power = 5000
        capacity = 10000
        config: ConfigParser = ConfigParser()
        electrolyte_system = VanadiumSystem(capacity, RedoxFlowConfig(config=config))
        if stack_module_typ_subclass.__name__ in [CellDataStack5500W.__name__, IndustrialStack1500W.__name__,
                                                  IndustrialStack9000W.__name__, HighPerformanceStack5700W.__name__]:
            return stack_module_typ_subclass(electrolyte_system, voltage, power, RedoxFlowDataConfig(),
                                             RedoxFlowConfig(config=config))
        else:
            return stack_module_typ_subclass(electrolyte_system, voltage, power, RedoxFlowConfig(config=config))
        #

    @pytest.fixture()
    def stack_module_subclass_list(self):
        # print(all_non_abstract_subclasses_of(StackModule))
        return all_non_abstract_subclasses_of(StackModule)

    def test_serial_scale_calculation(self, stack_module_subclass_list):
        for stack_module_subclass in stack_module_subclass_list:
            uut = self.make_stack_module(stack_module_subclass)
            # print(uut.get_parallel_scale(), uut.get_serial_scale())
            assert uut.get_parallel_scale() > 0
            assert uut.get_serial_scale() > 0

    def test_current_sign_check(self, stack_module_subclass_list):
        for stack_module_subclass in stack_module_subclass_list:
            uut = self.make_stack_module(stack_module_subclass)
            # print(uut.get_self_discharge_current(self.redox_flow_state))
            assert uut.get_self_discharge_current(self.redox_flow_state) > 0



from simses.commons.config.simulation.electrolyzer import ElectrolyzerConfig
from simses.commons.constants import Hydrogen
from simses.commons.state.technology.electrolyzer import ElectrolyzerState
from simses.technology.hydrogen.control.pressure.ideal_var_cathode import \
    IdealVarCathodePressureController
from simses.technology.hydrogen.electrolyzer.pressure.abstract_model import PressureModel
from simses.technology.hydrogen.electrolyzer.stack.stack_model import ElectrolyzerStackModel


class VarCathodePressureModel(PressureModel):

    def __init__(self, electrolyzer: ElectrolyzerStackModel, config: ElectrolyzerConfig):
        super().__init__()
        nominal_stack_power: float = electrolyzer.get_nominal_stack_power()
        self.VOLUME_GAS_SEPARATOR = 0.4225 * 10 ** (-3) * nominal_stack_power / 1000.0  # dm³ -> m³  0.4425 l/kW: value for Volume from H-TEC Series-ME: ME 450/1400
        self.__pressure_controller = IdealVarCathodePressureController(config)
        self.__p_c_1 = 0  # barg
        self.__p_a_1 = 0  # barg
        self.__n_h2_out_c = 0  # mol  mass which is set free through the control valve at cathode side
        self.__n_o2_out_a = 0  # mol  mass which is set free through the control valve at anode side
        self.__h2_outflow = 0  # mol/s
        self.__o2_outflow = 0  # mol/s
        self.__h2o_ouflow_cathode = 0  # mol/s
        self.__h2o_outflow_anode = 0  # mol/s

    def calculate(self, time, state: ElectrolyzerState):
        timestep = time - state.time
        stack_temp = state.temperature + 273.15  # °C -> K
        p_c_0 = state.pressure_cathode  # barg
        p_a_0 = state.pressure_anode  # barg
        p_h2o_0 = state.sat_pressure_h2o  # bar
        n_h2_prod = state.hydrogen_production  # mol/s at stack level
        x_h2o_c = p_h2o_0 / (1 + p_c_0 - p_h2o_0)
        n_o2_prod = state.oxygen_production  # mol/s at stack level
        x_h2o_a = p_h2o_0 / (1 + p_a_0 - p_h2o_0)
        pressure_factor_cathode = Hydrogen.IDEAL_GAS_CONST * stack_temp / self.VOLUME_GAS_SEPARATOR * \
                                  (1 + x_h2o_c) * 10 ** (-5)  # bar/mol   transfer from Pa to bar with 1bar 0 10^5 Pa = 10^5 N/m²
        self.__n_h2_out_c = self.__pressure_controller.calculate_n_h2_out(p_c_0, n_h2_prod, timestep, pressure_factor_cathode)
        self.__n_o2_out_a = self.__pressure_controller.calculate_n_o2_out(p_a_0, n_o2_prod, timestep)

        # new pressure cathode
        self.__p_c_1 = p_c_0 + pressure_factor_cathode * (n_h2_prod - self.__n_h2_out_c) * timestep  # bar

        # new pressure anode
        # self.p_a_1 = p_a_0 + ConstantsHydrogen.IDEAL_GAS_CONST * stack_temp / self.VOLUME_GAS_SEPARATOR * (1 + x_h2o_a) * \
        #              (n_o2_prod - self.n_o2_out_a) * 10 ** (-5)  # bar
        self.__p_a_1 = p_a_0

        # outflow rates of h2, o2 and h2o
        self.__h2_outflow = self.__n_h2_out_c  # mol/s
        self.__o2_outflow = self.__n_o2_out_a  # mol/s
        self.__h2o_ouflow_cathode = x_h2o_c * self.__h2_outflow  # mol/s
        self.__h2o_outflow_anode = x_h2o_a * self.__o2_outflow  # mol/s

    def get_pressure_cathode(self) -> float:
        return self.__p_c_1

    def get_pressure_anode(self) -> float:
        return self.__p_a_1

    def get_h2_outflow(self) -> float:
        return self.__h2_outflow

    def get_o2_outflow(self) -> float:
        return self.__o2_outflow

    def get_h2o_outflow_cathode(self) -> float:
        return self.__h2o_ouflow_cathode

    def get_h2o_outflow_anode(self) -> float:
        return self.__h2o_outflow_anode



from abc import ABC, abstractmethod
from datetime import datetime
import numpy
import pandas
from simses.analysis.utils import get_mean_for, get_positive_values_from, get_sum_for, get_negative_values_from, \
    get_min_for, get_max_for
from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.data.csv_data_handler import CSVDataHandler
from simses.commons.log import Logger


class Data(ABC):

    """
    Data is abstract class for providing time series values from the state values generated by SimSES. It provides
    many convenient methods to access simulation data in order to ease calculations in analysis.
    """

    __log: Logger = Logger(__name__)

    def __init__(self, config: GeneralSimulationConfig, data: pandas.DataFrame):
        self.__W_to_kWh = config.timestep / 3600 / 1000
        self.__data: pandas.DataFrame = data

    @property
    def convert_watt_to_kWh(self) -> float:
        """
        conversion coefficient in order to transform power in W to energy in kWh

        Returns
        -------
        float:
            conversion coefficient
        """
        return self.__W_to_kWh

    def _get_data(self, key: str) -> numpy.ndarray:
        """
        Returns data series for key as numpy array

        Parameters
        ----------
        key :

        Returns
        -------

        """
        return self.__data[key].to_numpy()

    def _get_first_value(self, key: str) -> float:
        """
        Returns fist value for data series of key

        Parameters
        ----------
        key :

        Returns
        -------

        """
        return float(self._get_data(key)[0])

    def _get_last_value(self, key: str) -> float:
        """
        Returns last value for data series of key

        Parameters
        ----------
        key :

        Returns
        -------

        """
        return float(self._get_data(key)[-1])

    def _get_difference(self, key: str) -> float:
        """

        Parameters
        ----------
        key :

        Returns
        -------

        """
        return self._get_last_value(key) - self._get_first_value(key)

    @property
    @abstractmethod
    def id(self) -> str:
        """

        Returns
        -------
        str:
            Data id as string
        """
        pass

    @property
    @abstractmethod
    def time(self):
        """
        Time series in s

        Returns
        -------

        """
        pass

    @property
    @abstractmethod
    def power(self):
        """
        Series of power values in W

        Returns
        -------

        """
        pass

    @property
    @abstractmethod
    def dc_power(self):
        """
        Series of power values in W

        Returns
        -------

        """
        pass

    @property
    def average_power(self) -> float:
        """
        Mean value of power series in W

        Returns
        -------

        """
        return get_mean_for(self.power)

    @property
    def charge_energy(self) -> float:
        """
        Charge energy in kWh (as positive values)

        Returns
        -------

        """
        return get_sum_for(self.charge_energy_series)

    @property
    def charge_energy_series(self) -> numpy.ndarray:
        """
        Series of charge energy in kWh (as positive values)

        Returns
        -------

        """
        return get_positive_values_from(self.power) * self.__W_to_kWh

    @property
    def discharge_energy(self) -> float:
        """
        Discharge energy in kWh (as positive values)

        Returns
        -------

        """
        return get_sum_for(self.discharge_energy_series)

    @property
    def discharge_energy_series(self) -> numpy.ndarray:
        """
        Series of discharge energy in kWh (as positive values)

        Returns
        -------

        """
        return -1 * get_negative_values_from(self.power) * self.__W_to_kWh

    @property
    def discharge_energy_per_year(self) -> numpy.ndarray:
        """
        Series of yearly discharge energy in kWh (as positive values)

        Returns
        -------

        """
        return self.__energy_per_year(self.discharge_energy_series)

    @property
    def charge_energy_per_year(self) -> numpy.ndarray:
        """
        Series of yearly charge energy in kWh (as positive values)

        Returns
        -------

        """
        return self.__energy_per_year(self.charge_energy_series)

    def __energy_per_year(self, energy: numpy.ndarray) -> numpy.ndarray:
        time: numpy.ndarray = self.time
        energy: numpy.ndarray = numpy.column_stack([time, energy])
        start: float = time[0]
        end: float = time[-1]
        slice_start: float = start
        energy_per_year: list = list()
        while slice_start <= end:
            date = datetime.fromtimestamp(slice_start)
            slice_end: float = date.replace(year=date.year + 1).timestamp()
            # 0: Column index for time series, 1: Column index for energy series
            yearly_energy = energy[numpy.where((energy[:, 0] >= slice_start) * (energy[:, 0] < slice_end))]
            energy_per_year.append(yearly_energy[:, 1].sum())
            slice_start = slice_end
        return numpy.array(energy_per_year)

    @property
    @abstractmethod
    def energy_difference(self):
        """
        Energy difference of start and end point in kWh

        Returns
        -------

        """
        pass

    @property
    @abstractmethod
    def soc(self):
        """
        Series of soc values in p.u.

        Returns
        -------

        """
        pass


    @property
    def average_soc(self):
        """
        Mean soc in p.u.

        Returns
        -------

        """
        return get_mean_for(self.soc)

    @property
    def max_soc(self):
        """
        Maximal soc of series in p.u.

        Returns
        -------

        """
        return get_max_for(self.soc)

    @property
    def min_soc(self):
        """
        Minimum soc of series in p.u.

        Returns
        -------

        """
        return get_min_for(self.soc)

    @property
    @abstractmethod
    def capacity(self):
        """
        Series of capacity in kWh

        Returns
        -------

        """
        pass

    @property
    @abstractmethod
    def state_of_health(self):
        """
        Series of state of health in p.u.

        Returns
        -------

        """
        pass

    @property
    @abstractmethod
    def storage_fulfillment(self):
        """
        Percentage of time the system or battery can fulfill the desired power

        Returns
        -------

        """
        pass

    @property
    def initial_capacity(self):
        """
        First value of capacity series in kWh

        Returns
        -------

        """
        return self.capacity[0]

    @property
    def initial_state_of_health(self):
        """
        First value of state of health series in p.u.

        Returns
        -------

        """
        return self.state_of_health[1]

    @classmethod
    @abstractmethod
    def get_system_data(cls, path: str, config: GeneralSimulationConfig) -> list:
        """
        Extracts unique systems data from storage data files in path

        Parameters
        ----------
        path : value folder
        config : simulation data_config in value folder

        Returns
        -------

        """
        pass

    @classmethod
    def _get_system_data_for(cls, path, state_cls, time_key: str, system_key: str = None, storage_key: str = None) -> [pandas.DataFrame]:
        """
        Extracts unique systems from storage data files

        Parameters
        ----------
        path : value folder
        state_cls : storage data state class
        system_key : Key for system id of storage data state class
        storage_key : Key for storage id of storage data state class
        time_key : Key for time of storage data state class

        Returns
        -------
            List of device data
        """
        # data_import: DataImport = DataImport(path, state_cls)
        # data: pandas.DataFrame = data_import.return_data()
        try:
            data: pandas.DataFrame = CSVDataHandler.get_data_from(path, state_cls)
            # return data without id selection
            if system_key is None or storage_key is None:
                data.sort_values(by=[time_key], inplace=True)
                return [data]
            # get unique systems
            systems: pandas.DataFrame = data[[system_key, storage_key]].copy()
            systems.drop_duplicates(inplace=True)
            # get data for unique systems
            res: [pandas.DataFrame] = list()
            for row in range(len(systems.index)):
                system = systems.iloc[row]
                system_id = system[system_key]
                storage_id = system[storage_key]
                system_data: pandas.DataFrame = data.loc[
                    (data[system_key] == system_id) & (data[storage_key] == storage_id)].copy()
                system_data.sort_values(by=[system_key, storage_key, time_key], inplace=True)
                if system_data.isna().values.any():
                    cls.__log.error('Data with system id ' + str(int(system_id)) + ' and storage id '
                                    + str(int(storage_id)) + ' has NaN values! This data will be neglected for analysis!')
                else:
                    res.append(system_data)
            return res
        except FileNotFoundError:
            cls.__log.info('File could not be found for ' + path + state_cls.__name__)
            return list()

    @staticmethod
    def sort_by_id(data: list, descending: bool = False) -> None:
        """
        In-place sorting of list with Data objects by id (ascending)

        Parameters
        ----------
        descending :
            reverse sorting of data list, default: False
        data :
            list of Data objects

        Returns
        -------

        """
        data.sort(key=lambda x: x.id, reverse=descending)

    @classmethod
    def close(cls):
        cls.__log.close()



from abc import ABC, abstractmethod


class MarketProfile(ABC):
    """
    Profile of market prices for a time frame.
    """
    def __init__(self):
        super().__init__()

    @abstractmethod
    def next(self, time: float) -> float:
        """
        provides the next market data (price) of a specific market

        Parameters
        ----------

        Returns
        -------
        float
            next price signal
        """
        pass

    @abstractmethod
    def initialize_profile(self) -> None:
        pass

    @abstractmethod
    def close(self):
        """
        Closing all open resources in market profile

        Parameters
        ----------

        Returns
        -------

        """
        pass



from configparser import ConfigParser

from simses.commons.log import Logger
from simses.commons.state.technology.hydrogen import HydrogenState
from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.config.simulation.hydrogen import HydrogenConfig
from simses.technology.hydrogen.control.management import HydrogenManagementSystem
from simses.technology.hydrogen.hydrogen_storage.hydrogen_storage import HydrogenStorage
from simses.technology.hydrogen.hydrogen_storage.pipeline.simple_pipeline import SimplePipeline
from simses.technology.hydrogen.hydrogen_storage.pressuretank.pressuretank import PressureTank


class HydrogenFactory:

    def __init__(self, config: ConfigParser):
        self.__log: Logger = Logger(type(self).__name__)
        self.__config_general: GeneralSimulationConfig = GeneralSimulationConfig(config)
        self.__config_hydrogen: HydrogenConfig = HydrogenConfig(config)

    def create_hydrogen_state_from(self, system_id: int, storage_id: int) -> HydrogenState:
        hs = HydrogenState(system_id, storage_id)
        hs.time = self.__config_general.start
        hs.soh = 1.0
        return hs

    def create_hydrogen_management_system(self, electrolyzer_minimal_power, electrolyzer_maximal_power: float, fuel_cell_maximal_power: float):
        return HydrogenManagementSystem(min_power_electroyzer=electrolyzer_minimal_power,
                                        max_power_electrolyzer=electrolyzer_maximal_power,
                                        max_power_fuel_cell=fuel_cell_maximal_power, config=self.__config_hydrogen)

    def create_hydrogen_storage(self, storage: str, capacity: float, max_pressure: float) -> HydrogenStorage:
        soc: float = self.__config_hydrogen.soc
        if storage == PressureTank.__name__:  # name of the file in which object pressuretank is located
            self.__log.debug('Creating pressuretank as ' + storage)
            return PressureTank(capacity, max_pressure, soc)
        elif storage == SimplePipeline.__name__:
            self.__log.debug('Creating pipeline as ' + storage)
            return SimplePipeline(storage_pressure=max_pressure)
        else:
            options: [str] = list()
            options.append(PressureTank.__name__)
            options.append(SimplePipeline.__name__)
            raise Exception('Specified pressuretank ' + storage + ' is unknown. '
                            'Following options are available: ' + str(options))

    def close(self):
        self.__log.close()



from abc import ABC

from simses.commons.config.simulation.battery import BatteryConfig
from simses.commons.cycle_detection.cycle_detector import CycleDetector
from simses.commons.error import EndOfLifeError
from simses.commons.log import Logger
from simses.commons.state.technology.lithium_ion import LithiumIonState
from simses.technology.lithium_ion.cell.type import CellType
from simses.technology.lithium_ion.degradation.calendar.calendar_degradation import \
    CalendarDegradationModel
from simses.technology.lithium_ion.degradation.cyclic.cyclic_degradation import \
    CyclicDegradationModel


class DegradationModel(ABC):
    """
    Model for the degradation behavior of the ESS by analysing the resistance increase and capacity decrease.
    """

    def __init__(self,
                 cell: CellType,
                 cyclic_degradation_model: CyclicDegradationModel,
                 calendar_degradation_model: CalendarDegradationModel,
                 cycle_detector: CycleDetector,
                 battery_config: BatteryConfig,
                 initial_degradation_possible: bool = False):
        super().__init__()
        self.__log: Logger = Logger(type(self).__name__)
        self.__calendar_degradation_model: CalendarDegradationModel = calendar_degradation_model
        self.__cyclic_degradation_model: CyclicDegradationModel = cyclic_degradation_model
        self.__cycle_detector: CycleDetector = cycle_detector
        self.__cell: CellType = cell
        self.__end_of_life: float = battery_config.eol
        self.__start_state_of_health: float = cell.get_soh_start()
        self.__total_capacity_loss: float = 0.0
        if self.__start_state_of_health < 1.0 and not initial_degradation_possible:
            raise Exception('\nThe degradation model for the cell type ' + type(cell).__name__ +
                            ' does not allow a start SOH < 1 (here START_SOH = ' + str(self.__start_state_of_health) + ').\n'
                            'Please change the cell type or select START_SOH = 1 in the config file. ')
        # Exception if start SOH is below EOL criterium
        if self.__start_state_of_health < self.__end_of_life:
            raise Exception('\nThe start SOH is below the EOL threshold. '
                            'Please choose a start SOH larger than the EOL threshold. Current values:'
                            + '\nSTART_SOH = ' + str(self.__start_state_of_health)
                            + '\nEOL = ' + str(self.__end_of_life))

    def update(self, time: float, battery_state: LithiumIonState) -> None:
        """
        Updating the capacity and resistance of the lithium_ion through the degradation model.

        Parameters
        ----------
        time : float
            Current timestamp.
        battery_state : LithiumIonState
            Current state of the lithium_ion.

        Returns
        -------

        """
        self.calculate_degradation(time, battery_state)

        # Capacity losses
        battery_state.capacity_loss_cyclic = self.__cyclic_degradation_model.get_degradation() \
                                             * battery_state.nominal_voltage
        battery_state.capacity_loss_calendric = self.__calendar_degradation_model.get_degradation() \
                                                * battery_state.nominal_voltage

        if battery_state.capacity_loss_cyclic != 0:
            battery_state.c_rate = self.__cycle_detector.get_crate() * 3600  # in 1/h
            battery_state.delta_dod = self.__cycle_detector.get_depth_of_cycle()  # in p.u.
        else:
            battery_state.c_rate = 0
            battery_state.delta_dod = 0

        dcapacity: float = 0.0
        dcapacity += battery_state.capacity_loss_cyclic
        dcapacity += battery_state.capacity_loss_calendric
        dcapacity += battery_state.capacity_loss_other

        self.__total_capacity_loss += dcapacity
        battery_state.capacity = self.__cell.get_capacity(battery_state) * self.__start_state_of_health * battery_state.nominal_voltage - self.__total_capacity_loss
        # Assumption: capacity losses lead to energy losses within battery, but SOC stays constant -> calculate new SOE
        # (recovery effects are neglected!)
        battery_state.soe = battery_state.soc * battery_state.capacity

        self.__log.debug('Capacity loss cyclic: ' + str(battery_state.capacity_loss_cyclic))
        self.__log.debug('Capacity loss calendric: ' + str(battery_state.capacity_loss_calendric))
        self.__log.debug('Capacity loss other: ' + str(battery_state.capacity_loss_other))
        self.__log.debug('Capacity loss total: ' + str(battery_state.capacity_loss_cyclic
                                                       + battery_state.capacity_loss_calendric))

        # Resistance increase
        battery_state.resistance_increase_cyclic = self.__cyclic_degradation_model.get_resistance_increase()
        battery_state.resistance_increase_calendric = self.__calendar_degradation_model.get_resistance_increase()
        battery_state.resistance_increase += (battery_state.resistance_increase_cyclic
                                              + battery_state.resistance_increase_calendric
                                              + battery_state.resistance_increase_other)

        self.__log.debug('Resistance increase cyclic: ' + str(battery_state.resistance_increase_cyclic))
        self.__log.debug('Resistance increase calendric: ' + str(battery_state.resistance_increase_calendric))
        self.__log.debug('Resistance increase other: ' + str(battery_state.resistance_increase_other))
        self.__log.debug('Resistance increase total: ' + str(battery_state.resistance_increase_cyclic
                                                             + battery_state.resistance_increase_calendric))

        battery_state.soh = battery_state.capacity / (self.__cell.get_capacity(battery_state) * battery_state.nominal_voltage)
        self.check_battery_replacement(battery_state)

        self.__log.debug('Capacity: ' + str(battery_state.capacity * battery_state.nominal_voltage))
        self.__log.debug('Resistance increase: ' + str(battery_state.resistance_increase))

    def calculate_degradation(self, time: float, battery_state: LithiumIonState) -> None:
        """
        Calculates the resistance increase and capacity decrease (calendar always and
        cyclic only, if a cycle was detected)

        Parameters
        ----------
        time : float
            Current timestamp.
        battery_state : LithiumIonState
            Current state of the lithium_ion.

        Returns
        -------
        """

        self.__calendar_degradation_model.calculate_degradation(time, battery_state)
        self.__calendar_degradation_model.calculate_resistance_increase(time, battery_state)
        # Cyclic Aging only if cycle is detected
        if self.__cycle_detector.cycle_detected(time, battery_state):
            self.__cyclic_degradation_model.calculate_degradation(battery_state)
            self.__cyclic_degradation_model.calculate_resistance_increase(battery_state)

    def check_battery_replacement(self, battery_state: LithiumIonState) -> None:
        """
        Checks eol criteria and replaces the battery has to be replaced if necessary

        Parameters
        ----------
        battery_state : LithiumIonState
            Current state of the lithium_ion.

        Returns
        -------

        """
        soh = battery_state.soh
        self.__log.debug('SOH: ' + str(soh * 100) + '%')
        if soh < self.__end_of_life:
            raise EndOfLifeError ('Battery SOH is below End of life criteria')
            # self.__log.info('Battery SOH is below End of life criteria (' + str(self.__end_of_life) +
            #                 '). Battery is replaced')
            # battery_state.capacity = self.__cell.get_capacity(battery_state) * battery_state.nominal_voltage
            # battery_state.resistance_increase = 0
            # # resets specific values within a the degradation models
            # self.__calendar_degradation_model.reset(battery_state)
            # self.__cyclic_degradation_model.reset(battery_state)
            # self.__cycle_detector.reset()

    def close(self) -> None:
        """
        Closing all resources in degradation model

        Returns
        -------

        """
        self.__log.close()
        self.__calendar_degradation_model.close()
        self.__cyclic_degradation_model.close()



from simses.technology.lithium_ion.cell.thermal.properties import ThermalCellProperties


class DefaultThermalCellProperties(ThermalCellProperties):

    __MIN_TEMPERATURE: float = 273.15  # K
    __MAX_TEMPERATURE: float = 333.15  # K
    __MASS: float = 0.05  # kg per cell
    __SPECIFIC_HEAT: float = 700  # J/kgK
    __CONVECTION_COEFFICIENT: float = 15  # W/m2K

    def __init__(self):
        super(DefaultThermalCellProperties, self).__init__(self.__MIN_TEMPERATURE, self.__MAX_TEMPERATURE, self.__MASS,
                                                           self.__SPECIFIC_HEAT, self.__CONVECTION_COEFFICIENT)



from configparser import ConfigParser

from simses.analysis.data.abstract_data import Data
from simses.analysis.data.electrolyzer import ElectrolyzerData
from simses.analysis.data.energy_management import EnergyManagementData
from simses.analysis.data.fuel_cell import FuelCellData
from simses.analysis.data.lithium_ion import LithiumIonData
from simses.analysis.data.redox_flow import RedoxFlowData
from simses.analysis.data.system import SystemData
from simses.analysis.evaluation.abstract_evaluation import Evaluation
from simses.analysis.evaluation.economic.economic_evaluation import EconomicEvaluation
from simses.analysis.evaluation.economic.revenue_stream.abstract_revenue_stream import RevenueStream
from simses.analysis.evaluation.economic.revenue_stream.demand_charge_reduction import DemandChargeReduction
from simses.analysis.evaluation.economic.revenue_stream.electricity_consumption import \
    ElectricityConsumptionRevenueStream
from simses.analysis.evaluation.economic.revenue_stream.energy_cost_reduction import EnergyCostReduction
from simses.analysis.evaluation.economic.revenue_stream.fcr_revenue_stream import FCRRevenue
from simses.analysis.evaluation.economic.revenue_stream.intraday_recharge import IntradayRechargeRevenue
from simses.analysis.evaluation.economic.revenue_stream.operation_and_maintenance import \
    OperationAndMaintenanceRevenue
from simses.analysis.evaluation.merger import EvaluationMerger
from simses.analysis.evaluation.technical.electrolyzer import ElectrolyzerTechnicalEvaluation
from simses.analysis.evaluation.technical.fuel_cell import FuelCellTechnicalEvaluation
from simses.analysis.evaluation.technical.lithium_ion import LithiumIonTechnicalEvaluation
from simses.analysis.evaluation.technical.redox_flow import RedoxFlowTechnicalEvaluation
from simses.analysis.evaluation.technical.site_level import SiteLevelEvaluation
from simses.analysis.evaluation.technical.system import SystemTechnicalEvaluation
from simses.commons.config.analysis.economic import EconomicAnalysisConfig
from simses.commons.config.analysis.general import GeneralAnalysisConfig
from simses.commons.config.analysis.market import MarketProfileConfig
from simses.commons.config.simulation.battery import BatteryConfig
from simses.commons.config.simulation.energy_management import EnergyManagementConfig
from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.config.simulation.system import StorageSystemConfig
from simses.commons.log import Logger
from simses.logic.energy_management.strategy.basic.frequency_containment_reserve import \
    FrequencyContainmentReserve
from simses.logic.energy_management.strategy.basic.intraday_market_recharge import \
    IntradayMarketRecharge
from simses.logic.energy_management.strategy.basic.peak_shaving_perfect_foresight import \
    PeakShavingPerfectForesight
from simses.logic.energy_management.strategy.basic.peak_shaving_simple import \
    SimplePeakShaving
from simses.logic.energy_management.strategy.basic.residential_pv_feed_in_damp import \
    ResidentialPvFeedInDamp
from simses.logic.energy_management.strategy.basic.residential_pv_greedy import \
    ResidentialPvGreedy
from simses.logic.energy_management.strategy.basic.use_all_renewable_energy import \
    UseAllRenewableEnergy
from simses.logic.energy_management.strategy.stacked.fcr_idm_recharge_stacked import \
    FcrIdmRechargeStacked


class AnalysisFactory:

    # TODO Is the dependence of simulation pkg acceptable? --> MM
    __ENERGY_COST_REDUCTION: [str] = [ResidentialPvGreedy.__name__, ResidentialPvFeedInDamp.__name__,
                                      SimplePeakShaving.__name__, PeakShavingPerfectForesight.__name__]
    __DEMAND_CHARGE_REDUCTION: [str] = [SimplePeakShaving.__name__, PeakShavingPerfectForesight.__name__]
    __FREQUENCY_CONTAINMENT_RESERVE: [str] = [FrequencyContainmentReserve.__name__, FcrIdmRechargeStacked.__name__]
    __INTRADAY_MARKET: [str] = [FcrIdmRechargeStacked.__name__, IntradayMarketRecharge.__name__]
    __ELECTRICITY_CONSUMPTION_REVENUE_STREAM: [str] = [UseAllRenewableEnergy.__name__]

    def __init__(self, path: str, config: ConfigParser, version: str):
        self.__log: Logger = Logger(type(self).__name__)
        self.__result_path: str = path
        self.__version: str = version
        self.__analysis_config: GeneralAnalysisConfig = GeneralAnalysisConfig(config)
        self.__economic_analysis_config: EconomicAnalysisConfig = EconomicAnalysisConfig(config)
        self.__market_profile_config: MarketProfileConfig = MarketProfileConfig(config)
        self.__simulation_config: GeneralSimulationConfig = GeneralSimulationConfig(config=None, path=self.__result_path)
        self.__battery_config: BatteryConfig = BatteryConfig(config=None, path=self.__result_path)
        self.__energy_management_config: EnergyManagementConfig = EnergyManagementConfig(config=None, path=self.__result_path)
        self.__storage_system_config: StorageSystemConfig = StorageSystemConfig(config=None, path=self.__result_path)
        self.__do_plotting: bool = self.__analysis_config.plotting
        self.__do_system_analysis: bool = self.__analysis_config.system_analysis
        self.__do_lithium_ion_analysis: bool = self.__analysis_config.lithium_ion_analysis
        self.__do_redox_flow_analysis: bool = self.__analysis_config.redox_flow_analysis
        self.__do_hydrogen_analysis: bool = self.__analysis_config.hydrogen_analysis
        self.__do_site_level_analysis: bool = self.__analysis_config.site_level_analysis
        self.__do_economic_analysis: bool = self.__analysis_config.economical_analysis
        try:
            self.__energy_management_data: EnergyManagementData = EnergyManagementData.get_system_data(self.__result_path, self.__simulation_config)[0]
        except IndexError:
            self.__log.warn('No energy management data found!')
            self.__energy_management_data = None

    def __create_data_list(self) -> [Data]:
        config = self.__simulation_config
        path = self.__result_path
        data_list: [Data] = list()
        data_list.extend(SystemData.get_system_data(path, config))
        data_list.extend(LithiumIonData.get_system_data(path, config))
        data_list.extend(RedoxFlowData.get_system_data(path, config))
        #  data_list.extend(HydrogenData.get_system_data(path, config))
        data_list.extend(ElectrolyzerData.get_system_data(path, config))
        data_list.extend(FuelCellData.get_system_data(path, config))
        for data in data_list:
            self.__log.info('Created ' + type(data).__name__)
        return data_list

    def __create_revenue_streams(self, system_data) -> [RevenueStream]:
        revenue_streams: [RevenueStream] = list()
        economic_config: EconomicAnalysisConfig = self.__economic_analysis_config
        energy_management_strategy: str = self.__energy_management_config.operation_strategy
        if self.__energy_management_data is None:
            self.__log.warn('No energy management data available ----> No economic evaluation possible!')
            return revenue_streams
        if energy_management_strategy in self.__ENERGY_COST_REDUCTION:
            revenue_streams.append(EnergyCostReduction(self.__energy_management_data, system_data, economic_config))
            self.__log.info('Adding ' + EnergyCostReduction.__name__ + ' to revenue streams')
        if energy_management_strategy in self.__DEMAND_CHARGE_REDUCTION:
            revenue_streams.append(DemandChargeReduction(self.__energy_management_data, system_data, economic_config))
            self.__log.info('Adding ' + DemandChargeReduction.__name__ + ' to revenue streams')
        if energy_management_strategy in self.__FREQUENCY_CONTAINMENT_RESERVE:
            revenue_streams.append(FCRRevenue(self.__energy_management_data, system_data, economic_config,
                                              self.__simulation_config, self.__market_profile_config))
            self.__log.info('Adding ' + FCRRevenue.__name__ + ' to revenue streams')
        if energy_management_strategy in self.__INTRADAY_MARKET:
            revenue_streams.append(IntradayRechargeRevenue(self.__energy_management_data, system_data, economic_config,
                                                           self.__simulation_config, self.__market_profile_config))
            self.__log.info('Adding ' + IntradayRechargeRevenue.__name__ + ' to revenue streams')
        if energy_management_strategy in self.__ELECTRICITY_CONSUMPTION_REVENUE_STREAM:
            revenue_streams.append(ElectricityConsumptionRevenueStream(self.__energy_management_data, system_data, economic_config))
            self.__log.info('Adding ' + ElectricityConsumptionRevenueStream.__name__ + ' to revenue streams')
        if not revenue_streams:
            self.__log.warn('No revenue streams are defined for chosen Energy Management Strategy: ' + energy_management_strategy)
        revenue_streams.append(OperationAndMaintenanceRevenue(self.__energy_management_data, system_data, economic_config))
        self.__log.info('Adding ' + OperationAndMaintenanceRevenue.__name__ + ' to revenue streams')
        return revenue_streams

    def create_evaluations(self) -> [Evaluation]:
        data_list: [Data] = self.__create_data_list()
        evaluations: [Evaluation] = list()
        config: GeneralAnalysisConfig = self.__analysis_config
        ems_config: EnergyManagementConfig = self.__energy_management_config
        path: str = self.__result_path
        for data in data_list:
            if isinstance(data, SystemData):
                if self.__do_system_analysis:
                    evaluations.append(SystemTechnicalEvaluation(data, config, path))
                if data.is_top_level_system and self.__energy_management_data is not None:
                    if self.__do_site_level_analysis:
                        evaluations.append(SiteLevelEvaluation(data, self.__energy_management_data, config, ems_config,
                                                               path))
                    if self.__do_economic_analysis:
                        revenue_streams = self.__create_revenue_streams(data)
                        evaluations.append(EconomicEvaluation(data, self.__economic_analysis_config, revenue_streams,
                                                              config, self.__storage_system_config))
            elif isinstance(data, LithiumIonData) and self.__do_lithium_ion_analysis:
                evaluations.append(LithiumIonTechnicalEvaluation(data, config, self.__battery_config, path))
            elif isinstance(data, RedoxFlowData) and self.__do_redox_flow_analysis:
                evaluations.append(RedoxFlowTechnicalEvaluation(data, config, path))
            #  elif isinstance(data, HydrogenData) and self.__do_hydrogen_analysis:
            #     evaluations.append(HydrogenTechnicalEvaluation(data, config, path))
            elif isinstance(data, ElectrolyzerData) and self.__do_hydrogen_analysis:
                evaluations.append(ElectrolyzerTechnicalEvaluation(data, config, path))
            elif isinstance(data, FuelCellData) and self.__do_hydrogen_analysis:
                evaluations.append(FuelCellTechnicalEvaluation(data, config, path))
        return evaluations

    def create_evaluation_merger(self) -> EvaluationMerger:
        return EvaluationMerger(self.__result_path, self.__analysis_config, self.__version)

    def close(self):
        self.__log.close()
        # TODO close Data but not for all instances
        # Data.close()



from abc import ABC, abstractmethod

from simses.commons.timeseries.timevalue import TimeValue


class Average(ABC):
    """
    Averages values of multiple TimeValues with a specific behaviour (arithmetic mean, median, ...)
    """

    def __init__(self):
        super().__init__()

    @abstractmethod
    def average(self, data: [TimeValue]) -> float:
        """
        Averages values of data

        Parameters
        ----------
        data :
            List of TimeValue objects

        Returns
        -------
        float:
            averaged value
        """
        pass



import numpy as np
import pandas as pd

from simses.commons.state.technology.electrolyzer import ElectrolyzerState
from simses.technology.hydrogen.electrolyzer.stack.pem_analytic.electrical_model import \
    PemElectricalModel
from simses.technology.hydrogen.electrolyzer.stack.pem_analytic.fluid_model import \
    PemFluidModel
from simses.technology.hydrogen.electrolyzer.stack.pem_analytic.membrane_model import \
    PemMembraneModel
from simses.technology.hydrogen.electrolyzer.stack.pem_analytic.pressure_model import \
    PemPressureModel


class PemElectrolyzerEfficiencyCurves:

    def __init__(self, electrical_model: PemElectricalModel, membrane_model: PemMembraneModel, pressure_model: PemPressureModel, fluid_modell: PemFluidModel):
        self.__electrical_model = electrical_model
        self.__membrane_model = membrane_model
        self.__pressure_model = pressure_model
        self.__fluid_model = fluid_modell

    def calculate_efficiency_curves(self, hydrogen_state: ElectrolyzerState, start=0.01, stop=3.01, step=0.01):
        # stack_temperature = hydrogen_state.temperature_el  # °C
        # p_anode = hydrogen_state.pressure_anode_el  # barg
        # p_cathode = hydrogen_state.pressure_cathode_el  # barg
        # resistance_increase = hydrogen_state.resistance_increase_el  # p.u.
        # exchange_current_decrease = hydrogen_state.exchange_current_decrease_el  # p.u.
        current_dens_series = pd.Series(np.arange(start, stop, step))
        current_cell_series = current_dens_series.multiply(self.__membrane_model.get_geometric_area_cell())  # cells are connected serial -> stack current = cell current
        voltage_cell_series = current_dens_series.apply(lambda x: self.__electrical_model.get_cell_voltage(x,
                                                                                            hydrogen_state))
        # voltage_stack_series = voltage_cell_series.apply(lambda x: x * self.__number_cells)
        rev_voltage = self.__electrical_model.get_rev_voltage(hydrogen_state.temperature)

        # calculation of partial pressures of hydrogen and oxygen and saturation pressure of water
        h2o_sat_pressure = self.__pressure_model.get_sat_pressure_h2o(hydrogen_state.temperature)
        part_pressure_h2_series = current_dens_series.apply(lambda x: self.__pressure_model.get_partial_pressure_h2(
                                                            hydrogen_state, x))
        part_pressure_o2_series = current_dens_series.apply(lambda x: self.__pressure_model.get_partial_pressure_o2(
                                                            hydrogen_state, x))

        # generation and use of water, hydrogen and oxygen
        h2_gen_cell_series = current_cell_series.apply(self.__fluid_model.get_h2_generation_cell)
        o2_gen_cell_series = current_cell_series.apply(self.__fluid_model.get_o2_generation_cell)
        h2o_use_cell_series = current_cell_series.apply(self.__fluid_model.get_h2o_use_cell)

        # permeation of hydrogen, oxygen and water through membrane
        # def clac_h2_perm_df(df):
        #     return self.calculate_h2_permeation_cell(df[0], df[1])
        # df_part_pressures = pd.DataFrame({'part pressure h2': part_pressure_h2_series,
        #                                   'part pressure o2': part_pressure_o2_series})
        # h2_perm_cell_series = df_part_pressures.apply(clac_h2_perm_df, axis=1)
        # o2_perm_cell_series = part_pressure_o2_series.apply(self.calculate_o2_permeation_cell)
        # h2o_perm_cell_series = current_cell_series.apply(self.calculate_h2o_permeation_cell)

        h2_perm_series = current_cell_series.apply(lambda x: self.__membrane_model.get_h2_permeation_for_cell(hydrogen_state, x))
        o2_perm_series = current_cell_series.apply(lambda x: self.__membrane_model.get_o2_permeation_for_cell(hydrogen_state, x))
        h2o_perm_series = current_cell_series.apply(self.__membrane_model.get_h2o_permeation_for_cell)

        # use of water, production of hydrogen and oxygen at cell level
        # def calc_h2_net_cat_df(df):
        #     return self.calculate_h2_net_cathode(df[0], df[1], df[2])
        # df_net_h2_cathode = pd.DataFrame({'h2 gen': h2_gen_cell_series, 'h2 perm': h2_perm_cell_series,
        #                                   'o2 perm': o2_perm_cell_series})
        h2_net_cathode_series = current_cell_series.apply(lambda x: self.__fluid_model.get_h2_net_cathode(hydrogen_state, x))
        h2_net_cathode_series[h2_net_cathode_series < 0] = np.nan  # replace negative values by nan, so that is not calculated in the faraday efficiency

        # efficiency curve
        voltage_efficiency_series = voltage_cell_series.apply(lambda x: rev_voltage / x)
        faraday_efficiency_series = h2_net_cathode_series.divide(h2_gen_cell_series)
        cell_efficiency_series = voltage_efficiency_series.multiply(faraday_efficiency_series)

        return pd.DataFrame({'current density': current_dens_series, 'cell efficiency': cell_efficiency_series,
                             'voltage efficiency': voltage_efficiency_series, 'faraday efficiency': faraday_efficiency_series})




import logging
from configparser import ConfigParser

from simses.commons.config.abstract_config import Config
from simses.commons.utils.utilities import get_path_for


class LogConfig(Config):

    CONFIG_NAME: str = 'logger'
    CONFIG_PATH: str = get_path_for(CONFIG_NAME + Config.DEFAULTS)
    __config_parser: ConfigParser = None

    def __init__(self, path: str = None):
        if path is None:
            path = self.CONFIG_PATH
        super().__init__(path, self.CONFIG_NAME, self.__config_parser)
        self.__section: str = 'LOGGING'

    @classmethod
    def set_config(cls, config: ConfigParser) -> None:
        cls.__config_parser = config

    @property
    def log_level(self) -> int:
        """Returns log level"""
        try:
            level: str = self.get_property(self.__section, 'LOG_LEVEL')
            if level == 'DEBUG':
                return logging.DEBUG
            elif level == 'INFO':
                return logging.INFO
            elif level == 'WARNING':
                return logging.WARNING
            elif level == 'ERROR':
                return logging.ERROR
            else:
                return logging.DEBUG
        except:
            return logging.ERROR

    @property
    def log_to_console(self) -> bool:
        return self.get_property(self.__section, 'LOG_TO_CONSOLE') in ['True']

    @property
    def log_to_file(self) -> bool:
        return self.get_property(self.__section, 'LOG_TO_FILE') in ['True']



import math
from abc import ABC, abstractmethod

from simses.commons.config.simulation.battery import BatteryConfig
from simses.commons.log import Logger
from simses.commons.state.technology.lithium_ion import LithiumIonState
from simses.commons.utils.utilities import format_float
from simses.technology.lithium_ion.cell.electric.properties import ElectricalCellProperties
from simses.technology.lithium_ion.cell.format.abstract import CellFormat
from simses.technology.lithium_ion.cell.thermal.properties import ThermalCellProperties


class CellType(ABC):
    """A CellType describes a generic lithium_ion cell type. Abstract
    methods have to be completed in inherited classes, e.g. LFP."""

    __RANGE: float = 0.1  # range for parameter search in p.u.
    __LARGE_NUMBER: float = 1e100

    def __init__(self, voltage: float, capacity: float, soh: float, electrical_props: ElectricalCellProperties,
                 thermal_props: ThermalCellProperties, cell_format: CellFormat, battery_config: BatteryConfig):
        super().__init__()
        self.__log: Logger = Logger(type(self).__name__)
        self.__exact_size: bool = battery_config.exact_size  # True if serial / parallel connection can be floats
        try:
            serial, parallel = battery_config.serial_scale, battery_config.parallel_scale
        except (TypeError, KeyError) as err:
            # print(err)
            serial, parallel = self.__nearest_neighbor(voltage, capacity, electrical_props.get_nominal_voltage(),
                                                       electrical_props.get_nominal_capacity())
        self.__SERIAL_SCALE: float = serial
        self.__PARALLEL_SCALE: float = parallel
        self.__consider_voltage_limit: bool = battery_config.consider_voltage_limit
        self.__soh_start: float = soh  # p.u.
        start_soh_share: float = battery_config.start_soh_share
        self.__calendar_capacity_loss_start = (1.0 - self.__soh_start) * start_soh_share  # p.u.
        self.__cyclic_capacity_loss_start = (1.0 - self.__soh_start) * (1.0 - start_soh_share)  # p.u.

        self.__calendar_resistance_increase_start = (1.0 - self.__soh_start) * start_soh_share * 5  # p.u.
        self.__cyclic_resistance_increase_start = (1.0 - self.__soh_start) * (1.0 - start_soh_share) * 5  # p.u.
        self.__resistance_increase_start = self.__calendar_resistance_increase_start + \
                                           self.__cyclic_resistance_increase_start

        self.__thermal_props: ThermalCellProperties = thermal_props
        self.__electrical_props: ElectricalCellProperties = electrical_props
        self.__cell_format: CellFormat = cell_format
        self.__log.debug('serial: ' + str(serial) + ', parallel: ' + str(parallel))

    def __nearest_neighbor(self, voltage: float, capacity: float, cell_voltage: float, cell_capacity: float):
        if self.__exact_size:
            serial: float = voltage / cell_voltage
            parallel: float = capacity / voltage / cell_capacity
            return serial, parallel
        elif voltage == 0 or capacity == 0:
            self.__log.debug('Target cell voltage or capacity are 0. Serial and parallel scale are set to '
                             'single cell (1,1)')
            return 1, 1
        # define search range
        min_serial: int = max(1, math.floor(voltage / cell_voltage * (1 - self.__RANGE)))
        max_serial: int = max(1, math.ceil(voltage / cell_voltage * (1 + self.__RANGE)))
        min_capacity: float = capacity * (1 - self.__RANGE)
        max_capacity: float = capacity * (1 + self.__RANGE)
        # set start parameters
        serial: int = max(1, round(voltage / cell_voltage))
        parallel: int = max(1, round(capacity / voltage / cell_capacity))
        res: float = cell_voltage * serial * cell_capacity * parallel
        # begin search
        for s in range(min_serial, max_serial, 1):
            voltage_tmp = cell_voltage * s
            parallel_min = math.floor(capacity / voltage_tmp / cell_capacity)
            parallel_max = math.ceil(capacity / voltage_tmp / cell_capacity)
            for p in range(parallel_min, parallel_max, 1):
                capacity_tmp = cell_capacity * p
                res_tmp = voltage_tmp * capacity_tmp
                if min_capacity < res_tmp < max_capacity and abs(res_tmp - capacity) < abs(res - capacity):
                    serial, parallel, res = s, p, res_tmp
                    self.__log.debug('Found better fit: ' + format_float(res) + ' Wh, ' + str(serial) + ', ' + str(parallel))
        # log results
        self.__log.debug('Difference:  ' + format_float(cell_voltage * serial - voltage) + 'V, '
                         + format_float((cell_voltage * serial - voltage) / voltage * 100.0) + '%')
        self.__log.debug('Difference: ' + format_float(cell_capacity * parallel - capacity/voltage) + 'Ah, '
                         + format_float((cell_capacity * parallel * voltage - capacity) / capacity * 100.0) + '%')
        self.__log.debug('Difference: ' + format_float(res - capacity) + 'Wh, '
                         + format_float((res - capacity) / capacity * 100.0) + '%')
        if res < min_capacity or res > max_capacity:
            self.__log.warn('System capacity is not in range of ' + format_float(self.__RANGE * 100.0) + '%')
        return serial, parallel

    def get_calendar_capacity_loss_start(self) -> float:
        return self.__calendar_capacity_loss_start

    def get_cyclic_capacity_loss_start(self) -> float:
        return self.__cyclic_capacity_loss_start

    def get_calendar_resistance_increase_start(self) -> float:
        return self.__calendar_resistance_increase_start

    def get_cyclic_resistance_increase_start(self) -> float:
        return self.__cyclic_resistance_increase_start

    def get_resistance_increase_start(self) -> float:
        return self.__resistance_increase_start

    def get_soh_start(self) -> float:
        return self.__soh_start

    def check_soc_range(self, soc: float) -> float:
        if soc < 0.0 or soc > 1.0:
            self.__log.warn(str(soc) + ' is out of interpolation range')
            soc = min(1.0, max(0.0, soc))
        return soc

    @abstractmethod
    def get_open_circuit_voltage(self, battery_state: LithiumIonState) -> float:
        """
        Determines the open circuit voltage based on the current BatteryState and a lookup table for every cell type.

        Parameters
        ----------
        battery_state : LithiumIonState
            Current state of the lithium_ion.

        Returns
        -------
        float:
            Open circuit voltage of the cell in V

        """
        pass

    @abstractmethod
    def get_internal_resistance(self, battery_state: LithiumIonState) -> float:
        """
        Determines the internal resistance based on the current BatteryState and a lookup table for every cell type.

        Parameters
        ----------
        battery_state : LithiumIonState
            Current state of the lithium_ion.

        Returns
        -------
        float:
            Internal resistance of the cell in Ohm

        """
        pass

    def get_coulomb_efficiency(self, battery_state: LithiumIonState) -> float:
        """
        Determines the coulomb efficiency based on if the lithium_ion is charging or discharging.

        Default: 1.0 -> losses are calculated via internal resistance

        Parameters
        ----------
        battery_state : LithiumIonState
            Current BatteryState of the lithium_ion. Used to determine if the lithium_ion is charging or discharging.

        Returns
        -------
        float:
            Coulomb efficiency value.
        """
        efficiency: float = self.__electrical_props.get_coulomb_efficiency()
        return efficiency if battery_state.is_charge else 1.0 / efficiency

    def get_nominal_voltage(self) -> float:
        """
        determines the nominal voltage of the given cell type.

        Returns
        -------
        float:
            Nominal voltage in V
        """
        return self.__electrical_props.get_nominal_voltage() * self.get_serial_scale()

    def get_min_temp(self) -> float:
        """
        determines the minimal allowed operation temperature of a given cell type

        Returns
        -------
        float:
            minimal operation temperature in  K
        """
        return self.__thermal_props.get_min_temperature()

    def get_max_temp(self) -> float:
        """
        determines the maximum allowed operation temperature of a given cell type

        Returns
        -------
        float:
            maximum operation temperature in K
        """
        return self.__thermal_props.get_max_temperature()

    def get_min_voltage(self) -> float:
        """
        determines the recommended minimal operation voltage of a battery cell according to data sheet

        Returns
        -------
        float:
            minimal operation voltage in V
        """
        return self.__electrical_props.get_min_voltage() * self.get_serial_scale()

    def get_max_voltage(self) -> float:
        """
        determines the recommended maximal operation voltage of a battery cell according to data sheet

        Returns
        -------
        float:
            maximum operation voltage in V
        """
        return self.__electrical_props.get_max_voltage() * self.get_serial_scale()

    def get_min_current(self, battery_state: LithiumIonState) -> float:
        """
        determines the minimal operation current of a battery cell
        Attention: the sign of current is defined in the convention that charging current is positive and discharging
        is negative, therefore the minimal value is defined with regard to the sign.

        Parameters
        ----------
        battery_state : LithiumIonState
            Current BatteryState of the lithium_ion. Used to determine if the lithium_ion is charging or discharging
            and to get the SOC and temperature

        Returns
        -------
        float:
            minimal current in A
        """
        c_rate_limit: float = self.get_nominal_capacity() * self.__electrical_props.get_max_discharge_rate()
        voltage_limit: float = self.__LARGE_NUMBER
        if self.__consider_voltage_limit:
            try:
                voltage: float = battery_state.voltage
                rint: float = battery_state.internal_resistance
                voltage_limit: float = (voltage - self.get_min_voltage()) / rint
            except ZeroDivisionError:
                self.__log.warn('A zero division error for the internal cell resistance (Rint = 0) has been caught.')
        return -max(0.0, min(c_rate_limit, voltage_limit))

    def get_max_current(self, battery_state: LithiumIonState) -> float:
        """
        determines the maximum operation current of a battery cell
        Attention: the sign of current is defined in the convention that charging current is positive and discharging
        is negative, therefore the minimal value is defined with regard to the sign.

        Parameters
        ----------
        battery_state : LithiumIonState
            Current BatteryState of the lithium_ion. Used to determine if the lithium_ion is charging or discharging
            and to get the SOC and temperature

        Returns
        -------
        float:
            maximum current in A
        """
        c_rate_limit: float = self.get_nominal_capacity() * self.__electrical_props.get_max_charge_rate()
        voltage_limit: float = self.__LARGE_NUMBER
        if self.__consider_voltage_limit:
            try:
                voltage: float = battery_state.voltage
                rint: float = battery_state.internal_resistance
                voltage_limit: float = (self.get_max_voltage() - voltage) / rint
            except ZeroDivisionError:
                self.__log.warn('A zero division error for the internal cell resistance (Rint = 0) has been caught.')
        return max(0.0, min(c_rate_limit, voltage_limit))

    def get_self_discharge_rate(self, battery_state: LithiumIonState) -> float:
        """
        determines the self-discharge rate of battery cell

        Returns
        -------
        float:
            self-discharge rate in Wh/s
        """
        return self.__electrical_props.get_self_discharge_rate() * battery_state.soc * self.get_serial_scale() \
               * self.get_parallel_scale()

    def get_nominal_capacity(self) -> float:
        """
        Determines the capacity of the battery under nominal conditions.

        Attention: depending on the battery topology here the capacity refers possibly to that of single cell, module,
        pack and so on.

        Returns
        -------
        float:
            battery capacity in Ah
        """
        return self.__electrical_props.get_nominal_capacity() * self.get_parallel_scale()

    def get_capacity(self, battery_state: LithiumIonState) -> float:
        """
        Determines the current capacity of the battery.

        Attention: depending on the battery topology here the capacity refers possibly to that of single cell, module,
        pack and so on.

        Parameters
        ----------
        battery_state : LithiumIonState
            Current BatteryState of the lithium-ion battery. Used to determine if the avaiable capacity depending on
            C-rate and temperature, if applicable for cell type.

        Returns
        -------
        float:
            battery capacity in Ah
        """
        return self.get_nominal_capacity()

    def get_name(self) -> str:
        """
        determines the class name of a cell  (E.g. SonyLFP)

        Returns
        -------
        str:
            class name of a cell
        """

        return type(self).__name__

    def get_serial_scale(self) -> float:
        """
        determines the number of serially connected battery cells

        Returns
        -------
        float:
            number of serially connected cells
        """
        return self.__SERIAL_SCALE

    def get_parallel_scale(self) -> float:
        """
        determines the number of parallelly connected battery cells

        Returns
        -------
        float:
            number of parallelly connected cells
        """
        return self.__PARALLEL_SCALE

    def get_mass(self) -> float:
        """
        determines the mass of battery cell

        Returns
        -------
        float:
            mass in kg
        """
        return self.__thermal_props.get_mass() * self.get_serial_scale() * self.get_parallel_scale()

    def get_surface_area(self) -> float:
        """
        determines the surface area of battery cell depending on the cell geometry

        Returns
        -------
        float:
            cell surface area in m^2
        """
        return self.__cell_format.get_surface_area() * self.get_serial_scale() * self.get_parallel_scale()

    def get_specific_heat(self) -> float:
        """
        determines the specific heat capacity of a battery cell

        Returns
        -------
        float:
            specific heat capacity in J/(kg*K)
        """
        return self.__thermal_props.get_specific_heat()

    def get_convection_coefficient(self) -> float:
        """
        determines the convective heat transfer coefficient of a battery cell

        Returns
        -------
        float:
            convective heat transfer coefficient in W/(m^2*K)
        """
        return self.__thermal_props.get_convection_coefficient()

    def get_volume(self) -> float:
        return self.__cell_format.get_volume() * self.get_serial_scale() * self.get_parallel_scale()

    @abstractmethod
    def close(self) -> None:
        pass



import os


def count_loc(lines):
    nb_lines = 0
    docstring = False
    for line in lines:
        line = line.strip()

        if line == "" \
           or line.startswith("#") \
           or docstring and not (line.startswith('"""') or line.startswith("'''"))\
           or (line.startswith("'''") and line.endswith("'''") and len(line) > 3)  \
           or (line.startswith('"""') and line.endswith('"""') and len(line) > 3):
            continue

        # this is either a starting or ending docstring
        elif line.startswith('"""') or line.startswith("'''"):
            docstring = not docstring
            continue

        else:
            nb_lines += 1

    return nb_lines


def count_loc_in_directory(directory: str, ext: str, start: str = '', consider_comments: bool = False):
    loc = 0
    #print(directory)
    for item in os.listdir(directory):
        filename = os.path.join(directory, item)
        if os.path.isfile(filename) and filename.endswith(ext) and os.path.basename(filename).startswith(start):
            #print(filename)
            with open(filename, 'r') as file:
                try:
                    lines = file.readlines()
                    if consider_comments:
                        loc += len(lines)
                    else:
                        loc += count_loc(lines)
                except UnicodeDecodeError:
                    print('Error in ' + filename)
        if os.path.isdir(filename):
            loc += count_loc_in_directory(filename, ext, start, consider_comments)
    return loc


path: str = os.getcwd() + '/../../'
print('Calculating lines of code for ' + path)
loc_with_documentation = count_loc_in_directory(path, '.py', '', True)
loc_with_tests = count_loc_in_directory(path, '.py', '')
loc_tests = count_loc_in_directory(path, '.py', 'test_')
print('LOC: ' + str(loc_with_tests - loc_tests), '| Test lines: ' + str(loc_tests),
      '| Not code lines: ' + str(loc_with_documentation - loc_with_tests), '| All: ' + str(loc_with_documentation))



from simses.commons.log import Logger
from simses.technology.hydrogen.fuel_cell.stack.stack_model import FuelCellStackModel


class NoFuelCell(FuelCellStackModel):
    """An No-Fuelcell is a special typ of a Fuelcell"""

    def __init__(self):
        super().__init__()
        self.__log: Logger = Logger(type(self).__name__)

    def calculate(self, power: float):
        self.__voltage = 1  # V
        self.__current = 0  # A
        self.__hydrogen_consumption = 0  # mol/s

    def get_current(self):
        #print('the current is: ' + str(self.__current) + ' mA')
        return self.__current

    def get_hydrogen_consumption(self):
        #print('the massflow of hydrogen ist: ' + str(self.__mass_flow_h2) + ' mol/s')
        return self.__hydrogen_consumption

    def get_voltage(self):
        #print('the voltage is: ' + str(self.__voltage) + ' V')
        return self.__voltage

    def get_nominal_stack_power(self):
        return 0

    def get_number_cells(self):
        return 0

    def get_geom_area_stack(self):
        return 1

    def get_heat_capactiy_stack(self):
        return 0

    def close(self):
        self.__log.close()




from simses.commons.cycle_detection.cycle_detector import CycleDetector
from simses.commons.log import Logger
from simses.commons.state.technology.lithium_ion import LithiumIonState
from simses.technology.lithium_ion.cell.type import CellType
from simses.technology.lithium_ion.degradation.cyclic.cyclic_degradation import CyclicDegradationModel

import numpy as np


class LFPGenericCyclicDegradationModel(CyclicDegradationModel):

    def __init__(self, cell_type: CellType, cycle_detector: CycleDetector):
        super().__init__(cell_type, cycle_detector)
        self.__capacity_loss = 0
        self.__initial_capacity = self._cell.get_nominal_capacity()
        self.__log: Logger = Logger(type(self).__name__)

        # parameters
        self.__z = 0.392611936548533
        self.__a = 1.06456066666941
        self.__b = -10931.5112862013
        self.__c = 1
        self.__d = 1
        self.__e = -4.11988721392584
        self.__f = 1.70180851471426
        self.__g = 1.81362732352495
        self.__h = -0.217897784225967
        self.__i = 0.0201876349197979
        self.__j = -0.0382242087224787
        self.__k = 0.0185488829889972
        self.__l = 0.00326042032279996
        self.__m = 0.313675037732425
        self.__n = 1.90274120973485

        # constants
        self.__R = 8.314462  # universal gas constant

    def fec_dependency(self, fec):
        # degradation depends on fec as a power law function
        return fec**self.__z

    def temperature_dependency(self, temp):
        # degradation depends on temperature as arrhenius function
        return self.__a * np.exp(- self.__b / (self.__R * temp))

    def mean_soc_dependency(self, msoc):
        # degradation depends on msoc as a cubic function
        # limit msoc to avoid negative function values
        # implement warnings
        if msoc > 0.75:
            msoc = 0.75
            self.__log.warn('Mean state of charge for detected cycle was above 75%. MSoC was set to 75% to avoid '
                            'negative degradation.')
        elif msoc < 0.25:
            msoc = 0.25
            self.__log.warn('Mean state of charge for detected cycle was below 25%. MSoC was set to 25% to avoid '
                            'negative degradation.')
        return self.__e * msoc**3 + self.__f * msoc**2 + self.__g * msoc + self.__h

    def doc_dependency(self, doc):
        # degradation depends on doc as a cubic function
        return self.__i * doc**3 + self.__j * doc**2 + self.__k * doc + self.__l

    def crate_dependency(self, crate):
        # degradation depends on crate as a linear function
        return self.__m * crate + self.__n

    def calculate_degradation(self, battery_state: LithiumIonState) -> None:
        c_rate: float = self._cycle_detector.get_crate() * 3600
        depth_of_discharge: float = self._cycle_detector.get_depth_of_cycle()
        full_equivalent_cycle: float = self._cycle_detector.get_full_equivalent_cycle()
        mean_soc: float = self._cycle_detector.get_mean_soc()
        temperature: float = battery_state.temperature
        delta_fec: float = self._cycle_detector.get_delta_full_equivalent_cycle()

        # calculate degradation in percent (divide by 100 to get p.u.) with function 11 from MA Sandner
        capacity_loss_after = self.fec_dependency(full_equivalent_cycle) * \
                              self.temperature_dependency(temperature) * \
                              self.mean_soc_dependency(mean_soc) * \
                              self.doc_dependency(depth_of_discharge) * \
                              self.crate_dependency(c_rate)
        capacity_loss_before = self.fec_dependency(full_equivalent_cycle - delta_fec) * \
                               self.temperature_dependency(temperature) * \
                               self.mean_soc_dependency(mean_soc) * \
                               self.doc_dependency(depth_of_discharge) * \
                               self.crate_dependency(c_rate)

        self.__capacity_loss = (capacity_loss_after - capacity_loss_before) / 100.0 * self.__initial_capacity

    def calculate_resistance_increase(self, battery_state: LithiumIonState) -> None:
        # no resistance increase calculated for generic degradation model
        pass

    def get_degradation(self) -> float:
        degradation = self.__capacity_loss
        self.__capacity_loss = 0
        return degradation

    def get_resistance_increase(self) -> float:
        # no resistance increase calculated for generic degradation model
        return 0

    def reset(self, lithium_ion_state: LithiumIonState) -> None:
        self.__capacity_loss = 0

    def close(self) -> None:
        pass



import numpy as np

from simses.commons.log import Logger
from simses.commons.utils.utilities import check
from simses.system.power_electronics.acdc_converter.abstract_acdc_converter import AcDcConverter


class NottonLossAcDcConverter(AcDcConverter):

    # Notton Type 2 inverter
    __P0 = 0.0072
    __K = 0.0345

    __VOLUMETRIC_POWER_DENSITY = 143 * 1e6  # W / m3
    __GRAVIMETRIC_POWER_DENSITY = 17000  # W/kg
    __SPECIFIC_SURFACE_AREA = 0.0001  # in m2 / W  # TODO add exact values
    # Exemplary value from:
    # (https://www.iisb.fraunhofer.de/en/research_areas/vehicle_electronics/dcdc_converters/High_Power_Density.html)
    # ( https://www.apcuk.co.uk/app/uploads/2018/02/PE_Full_Pack.pdf )

    def __init__(self, max_power):
        super().__init__(max_power)
        self.__log: Logger = Logger(type(self).__name__)

    def to_ac(self, power: float, voltage: float) -> float:
        check(power)
        res: float = 0.0
        if power < 0.0:
            res = min(power - self.__get_loss(power), 0)
        return res

    def to_dc(self, power: float, voltage: float) -> float:
        check(power)
        res: float = 0.0
        if power > 0.0:
            res = max(power - self.__get_loss(power), 0)
        return res

    def to_dc_reverse(self, dc_power: float, voltage: float) -> float:
        check(dc_power)
        res: float = 0.0
        if dc_power > 0.0:
            p = - dc_power / (1 - dc_power * self.__K / self.max_power)
            q = - self.__P0 * dc_power / (1 / self.max_power - abs(dc_power) * self.__K / self.max_power ** 2)
            res = max(0.0, -p / 2 + np.sqrt((p / 2) ** 2 - q))
        return res

    def to_ac_reverse(self, dc_power: float, voltage: float) -> float:
        check(dc_power)
        res: float = 0.0
        if dc_power < 0.0:
            p = self.max_power / self.__K
            q = (self.__P0 * self.max_power ** 2 - abs(dc_power) * self.max_power) / self.__K
            self.__log.debug('P_DC: ' + str(dc_power))
            res = min(0.0, -(-p / 2 + np.sqrt((p / 2) ** 2 - q)))
        return res

    def __get_loss(self, power: float) -> float:
        return abs(self.__P0 * self.max_power + (self.__K / self.max_power) * power ** 2)

    @property
    def volume(self) -> float:
        return self.max_power / self.__VOLUMETRIC_POWER_DENSITY

    @property
    def mass(self):
        return self.max_power / self.__GRAVIMETRIC_POWER_DENSITY

    @property
    def surface_area(self) -> float:
        return self.max_power * self.__SPECIFIC_SURFACE_AREA

    @classmethod
    def create_instance(cls, max_power: float, power_electronics_config=None):
        return NottonLossAcDcConverter(max_power)

    def close(self) -> None:
        pass



from simses.logic.thermal_management.thermal_management import ThermalManagement
from simses.system.auxiliary.heating_ventilation_air_conditioning.fan import Fan
from simses.system.thermal.model.zero_d_system_thermal_model import ZeroDSystemThermalModel
from numpy import sign
import pandas as pd
import os
import time


class OnOffController(ThermalManagement):

    TEMPERATURE_DEAD_BAND: float = 2.0  # in K

    def __init__(self, set_point: float, hvac_max_thermal_power: float):
        super().__init__()
        self.__set_temperature = set_point  # K
        self.__hvac_rated_thermal_power = hvac_max_thermal_power  # W
        self.__export = False

        # Export running data
        if self.__export:
            self.__file_name = os.getcwd() + '\\controller_running_data\\Results' + str(time.time()) + '.csv'
            df = pd.DataFrame(['Ti', 'Ta', 'Total Thermal Power', 'HVAC Thermal Power', 'Required thermal power', 'HVAC Limit']).T
            df.to_csv(self.__file_name, header=None, index=None)

    def compute(self, internal_air_temperature: float, ambient_air_temperature: float, air_mass: float,
                air_density: float, time_step: float, fan: Fan) -> list:

        if abs(internal_air_temperature - self.__set_temperature) > self.TEMPERATURE_DEAD_BAND:
            required_thermal_power = ZeroDSystemThermalModel.AIR_SPECIFIC_HEAT * (internal_air_temperature - self.__set_temperature) * air_mass / time_step  # W
            rated_supply_temperature = float(self.__set_temperature - sign(required_thermal_power) * 10)  # K
            rated_fresh_air_thermal_power = fan.rated_airflow * air_density * (internal_air_temperature - ambient_air_temperature) * ZeroDSystemThermalModel.AIR_SPECIFIC_HEAT  # W
            max_physical_fresh_air_thermal_power = air_mass * (internal_air_temperature - ambient_air_temperature) * ZeroDSystemThermalModel.AIR_SPECIFIC_HEAT / time_step  # W

            if sign(required_thermal_power) == sign(rated_fresh_air_thermal_power):
                # if the fresh air thermal power is usable
                fresh_air_thermal_power = float(min(abs(rated_fresh_air_thermal_power), abs(max_physical_fresh_air_thermal_power)) * sign(rated_fresh_air_thermal_power))
                if (fresh_air_thermal_power / required_thermal_power) >= 0.999 or (ambient_air_temperature - self.__set_temperature) / (rated_supply_temperature - self.__set_temperature) >= 0.999:
                    # first case: fresh air thermal power is enough
                    # or fresh air and HVAC cannot be working at the same time, and fresh air thermal power is larger
                    # example: set point 298.15K, HVAC supply 288.15K, ambient 280K
                    hvac_thermal_power = 0.0  # W
                else:
                    # second case: the rest of demand from HVAC
                    hvac_thermal_power = float(min(abs(self.__hvac_rated_thermal_power), abs(ZeroDSystemThermalModel.AIR_SPECIFIC_HEAT * (ambient_air_temperature - rated_supply_temperature) * fan.rated_airflow * air_density)) * sign(required_thermal_power))  # W
            else:
                # if the fresh air thermal power cannot be used
                # third case: HVAC only
                fresh_air_thermal_power = 0.0
                hvac_thermal_power = float(min(abs(self.__hvac_rated_thermal_power), abs(ZeroDSystemThermalModel.AIR_SPECIFIC_HEAT * (internal_air_temperature - rated_supply_temperature) * fan.rated_airflow * air_density)) * sign(required_thermal_power))  # W

            total_thermal_power = float(hvac_thermal_power + fresh_air_thermal_power)  # W
            fan.run(fan.rated_airflow)

            # Export running data
            if self.__export:
                data_list = [float(internal_air_temperature), ambient_air_temperature, total_thermal_power,
                         hvac_thermal_power, required_thermal_power,
                         self.__hvac_rated_thermal_power]
                df = pd.DataFrame(data_list).T
                df.to_csv(self.__file_name, mode='a', header=None, index=None)

            return [total_thermal_power, hvac_thermal_power]
        else:
            fan.run(0)
            return [0.0, 0.0]



from abc import abstractmethod

from simses.system.auxiliary.auxiliary import Auxiliary

class Compressor(Auxiliary):
    """ Compressor is an auxiliary. It calculates the necessary electric power in W which is needed to compress an ideal gas
    from pressure 1 to pressure 2"""

    def __init__(self):
        super().__init__()

    def ac_operation_losses(self) -> float:
        """
        Calculates the ac operation losses

        Parameters
        ----------

        :returns
        -------
        float:
            Power in W
        """
        return self.get_compression_power()

    @abstractmethod
    def calculate_compression_power(self, hydrogen_flow_out:float, pressure_1: float, pressure_2: float, temperature: float) -> None:
        """
        Calculates the compression power

        :returns
        -------

        """
        pass

    @abstractmethod
    def get_compression_power(self) -> float:
        """
        Gets compressor power in W

        :returns
        --------

        """
        pass



"""
    This script uses relationships proposed in:

    "New multi-physics approach for modelling and design of alkaline
    electrolysers" by M. Hammoudi et al.

    "Simulation tool based on a physics model and an electrical analogy for an alkaline electrolyser" by C. Henao et al.

    to generate the polarization curve of an alkaline elektrolyser, depending on:
    - temperature
    - pressure
    - electrolyte concentration
    as well as other parameters of the chosen specific electrolyser (In this case the Stuart HRI Electrolyser example was used)

    The resulting curves for current density and power are saved in .csv files, to later be imported by the alkaline
    electrolyser stack model for usage in SimSES.
"""

import glob
import math as m

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

# Parameters TODO more general approach for other Electrolysers (here: HRI)
koh_weight_concentration = 30                           # %
anode_surface = 0.03                                    # m^2
cathode_surface = 0.03                                  # m^2
anode_thickness = 0.002                                 # m
cathode_thickness = 0.002                               # m
anode_membrane_gap = 0.00125                            # m
cathode_membrane_gap = 0.00125                          # m
membrane_surface_area = 0.03                            # m^2
limiting_current_density = 300000                       # A/m^2
standard_temperature = 25                               # °C (Room Temperature)
standard_temperature = standard_temperature + 273.15    # K

# Input Variables
temperatures = [30.0, 40.0, 50.0, 60.0, 70.0, 80.0, 90.0, 100.0, 110.0, 120.0, 130.0, 140.0, 150.0, 160.0]  # °C
pressures = np.linspace(1, 50, 30)                # bar
currents = np.linspace(0, 150, 300)

lookup_table_activation_overpotential = pd.read_csv('activation_overpotential_lookup_alkaline.csv', compression='gzip')

__IDEAL_GAS_CONST: float = 8.314462  # J/(mol K)
__FARADAY_CONST: float = 96485.3321  # As/mol

fitting_parameters = None

def calculate_cell_voltage(current: float, temperature: float, pressure: float) -> float:
    v_cell = 0

    # Convert temperature from °C to K
    temperature = (temperature + 273.15)

    # Get theoretical cell voltage
    v_theoretical = calculate_theoretical_voltage(temperature, pressure)
    v_cell += v_theoretical

    if current >= 0:
        # Get electrode activation overvoltages
        v_act_a = calculate_activation_overvoltage_anode(current, temperature)
        v_act_c = calculate_activation_overvoltage_cathode(current, temperature)
        v_activation_total = v_act_a + v_act_c
        v_cell += v_activation_total
        # Calculate ohmic resistances
        # Electrolyte
        v_electrolyte_resistance = calculate_electrolyte_overvoltage(current, temperature)
        v_cell += v_electrolyte_resistance
        # Electrodes
        v_electrode_resistances = calculate_electrode_overvoltage(current, temperature)
        v_cell += v_electrode_resistances
        # Membrane
        v_membrane_resistance = calculate_membrane_overvoltage(current, temperature)
        v_cell += v_membrane_resistance

    return v_cell


def calculate_theoretical_voltage(temperature: float, pressure: float) -> float:
    # Thermodynamic approximation of the reversible voltage at standard conditions
    v_reversible_standard_conditions = 1.50342 - 9.956 * 10 ** (-4) * temperature + 2.5 * 10 ** (-7) * temperature ** 2

    molarity = calculate_molarity(temperature)

    # Calculate vapour pressure of pure water pw* and pressure of gaseous KOH solution pw
    # (From: "The Thermodynamics of Aqueous Water Electrolysis" by LeRoy)
    pure_water_vapour_pressure = (temperature ** -3.4159) * np.exp(37.043 - (6275.7 / temperature))
    koh_solution_pressure = (temperature ** -3.498) * np.exp(37.93 - (6426.32 / temperature)) * np.exp(0.016214 - 0.13802 * molarity + 0.19330 * molarity**0.5)
    # Calculate water electrolysis theoretical voltage
    v_theoretical = v_reversible_standard_conditions + (
                (__IDEAL_GAS_CONST * temperature) / (2 * __FARADAY_CONST)) * np.log(
        (pressure - koh_solution_pressure) ** (3 / 2) * (pure_water_vapour_pressure / koh_solution_pressure))

    # Correction terms for the consideration of real gas instead of ideal gas
    v_correction1 = (pressure - koh_solution_pressure) * (21.661 * 10 ** (-6) - (5.471 * 10 ** (-3)) / temperature)
    v_theoretical += v_correction1
    v_correction2 = (pressure - koh_solution_pressure) ** 2 * (
                -1 * ((6.289 * 10 ** -6) / temperature) + ((0.135 * 10 ** -3) / temperature ** 1.5)
                + ((2.547 * 10 ** -3) / temperature ** 2) - (0.4825 / temperature ** 3))
    v_theoretical += v_correction2

    return v_theoretical


def calculate_activation_overvoltage_anode(current: float, temperature: float) -> float:
    # Calculate bubble rate coverage of electrodes (assumed to be identical for both electrodes)
    bubble_rate_coverage = calculate_bubble_rate_coverage(current, temperature)
    # Calculate nominal electrode surfaces
    anode_surface_nom = anode_surface * (1 - bubble_rate_coverage)
    current_density = current / (anode_surface_nom * 10000.0) # A/cm2
    # Calculate transfer coefficient for anode and cathode for HRI Electrolyser TODO more general approach?
    alpha_a = 0.0675 + 0.00095 * temperature
    # Calculate tafel slope for anode and cathode
    tafel_slope = (__IDEAL_GAS_CONST * temperature) / (2*__FARADAY_CONST * alpha_a)
    # Calculate exchange current densities for anode and cathode for HRI Electrolyser (Ni Electrodes) TODO more general approach?
    exchange_current_density = 30.4 - 0.206 * temperature + 0.00035 * temperature ** 2  # mA/cm2 ?
    # Calculate activation overvoltages of the electrodes:
    v_activation_anode = tafel_slope * m.asinh(current_density / (2*exchange_current_density / 1000.0))

    return v_activation_anode


def calculate_activation_overvoltage_cathode(current: float, temperature: float) -> float:
    # Calculate bubble rate coverage of electrodes (assumed to be identical for both electrodes)
    bubble_rate_coverage = calculate_bubble_rate_coverage(current, temperature)
    # Calculate nominal electrode surfaces
    cathode_surface_nom = cathode_surface * (1 - bubble_rate_coverage)
    current_density = current / (cathode_surface_nom * 10000.0)
    # Calculate transfer coefficient for anode and cathode for HRI Electrolyser TODO more general approach?
    alpha_c = 0.1175 + 0.00095 * temperature
    # Calculate tafel slope for anode and cathode
    tafel_slope = (__IDEAL_GAS_CONST * temperature) / (2*__FARADAY_CONST * alpha_c)
    # Calculate exchange current densities for anode and cathode for HRI Electrolyser (Ni Electrodes) TODO more general approach?
    exchange_current_density = 13.72491 - 0.09055 * temperature + 0.09055 * temperature ** 2  # mA/cm2 ?
    # Calculate activation overvoltages of the electrodes:
    v_activation_cathode = tafel_slope * m.asinh(current_density / (2*exchange_current_density / 1000.0))

    return v_activation_cathode


def calculate_activation_overvoltage_lookup(current: float, temperature: float) -> float:
    bubble_rate_coverage = calculate_bubble_rate_coverage(current, temperature)
    electrode_surface_nom = anode_surface * (1 - bubble_rate_coverage)
    current_density = current / (electrode_surface_nom * 10000.0)

    temperature = temperature - 273.15  # °C
    temperature_idx = abs(lookup_table_activation_overpotential['temperature in °C'] - temperature).idxmin()
    filter_temperature = lookup_table_activation_overpotential.iloc[temperature_idx, 0]  # °C
    filtered_table = lookup_table_activation_overpotential[lookup_table_activation_overpotential['temperature in °C'] == filter_temperature]
    filtered_table.reset_index(drop=True, inplace=True)

    current_idx = abs(filtered_table['current density in A/cm2'] - current_density).idxmin()
    activation_overvoltage = filtered_table.iloc[current_idx, 2] + filtered_table.iloc[current_idx, 3]

    return activation_overvoltage


def calculate_electrode_overvoltage(current: float, temperature: float) -> float:
    # Calculate electrical conductivity of Ni TODO more general approach?
    conductivity_nickel = 14.3 * 10**6  # S/m generic value for Nickel at room temperature TODO find good source
    conductivity_nickel = conductivity_nickel / 100   # S/cm
    # Calculate electrode resistances
    resistance_anode = (1 / conductivity_nickel) * ((anode_thickness * 100) / (anode_surface * 10000))
    resistance_cathode = (1 / conductivity_nickel) * ((cathode_thickness * 100) / (cathode_surface * 10000))
    v_electrode_resistances = (resistance_anode + resistance_cathode) * current

    return v_electrode_resistances


def calculate_electrolyte_overvoltage(current: float, temperature: float) -> float:
    molarity = calculate_molarity(temperature)
    bubble_rate_coverage = calculate_bubble_rate_coverage(current, temperature)

    # Calculate ionic conductivity of KOH in S/cm
    conductivity_koh_free = -2.041 * molarity - 0.0028 * molarity**2 + 0.005332 * molarity * temperature + 207.2 * (
                molarity / temperature) \
                     + 0.001043 * molarity**3 - 0.0000003 * molarity**2 * temperature**2
    # Calculate free resistance of electrolyte
    resistance_electrolyte_free = (1 / conductivity_koh_free) * ((anode_membrane_gap / anode_surface + cathode_membrane_gap / cathode_surface) / 100)
    # Incorporate bubble rate into resistance
    resistance_electrolyte_from_bubble_rate = resistance_electrolyte_free * ((1 / ((1 - (2 / 3) * bubble_rate_coverage) ** (3 / 2))) - 1)
    resistance_electrolyte_total = resistance_electrolyte_free + resistance_electrolyte_from_bubble_rate
    v_electrolyte_resistance = resistance_electrolyte_total * current

    return v_electrolyte_resistance


def calculate_membrane_overvoltage(current: float, temperature: float) -> float:
    # Calculate Membrane resistance (HRI Electrolyser: 0.5mm Zirfon) TODO more general approach?
    # resistance_membrane = (0.060 + 80 * m.exp(temperature / 50)) / (10000 * membrane_surface_area * 10000)

    # Generic values from "VERMEIREN, P. (1998). Evaluation of the ZirfonS separator for use in alkaline water electrolysis and Ni-H2 batteries.
    # International Journal of Hydrogen Energy, 23(5), 321–324. doi:10.1016/s0360-3199(97)00069-4" (Regression from graph)
    resistance_membrane_per_area = fitting_parameters.zirfon1 * temperature ** 2 + fitting_parameters.zirfon2 * temperature + fitting_parameters.zirfon3
    resistance_membrane = resistance_membrane_per_area / (membrane_surface_area * 10000)
    v_membrane_resistance = resistance_membrane * current

    return v_membrane_resistance


def calculate_molarity(temperature: float) -> float:
    # Calculate molarity of KOH electrolyte
    # From "New multi-physics approach for modelling and design of alkaline electrolysers" by M. Hammoudi et al.
    molarity = (koh_weight_concentration * (
            183.1221 - 0.56845 * temperature + 984.5679 * m.exp(koh_weight_concentration / 115.96277))) / (
                       100 * 56.105)

    return molarity


def calculate_bubble_rate_coverage(current: float, temperature: float) -> float:
    bubble_rate_coverage = (-97.25 + 182 * (temperature / standard_temperature) - 84 * (temperature / standard_temperature) ** 2) * (
            (current / cathode_surface) / limiting_current_density) ** 0.3

    return bubble_rate_coverage


def generate_polarisation_lookup():
    lookup_voltages = []
    lookup_temperatures = []
    lookup_pressures = []
    lookup_currents = []

    for i in range(len(temperatures)):
        for p in range(len(pressures)):
            for j in range(len(currents)):
                # lookup_voltages.append(calculate_cell_voltage(currents[j], temperatures[i], pressures[p]))
                xdata = np.array([[temperatures[i]], [pressures[p]], [currents[j]]])
                lookup_voltages.append(cell_voltage_fit_model_function(xdata, fitting_parameters.ac, fitting_parameters.bc, fitting_parameters.cc,
                                                        fitting_parameters.aa, fitting_parameters.ba, fitting_parameters.ca,
                                                        fitting_parameters.r1, fitting_parameters.r2, fitting_parameters.r3,
                                                        fitting_parameters.a1, fitting_parameters.a2)[0])
                lookup_temperatures.append(temperatures[i])
                lookup_pressures.append(pressures[p])
                lookup_currents.append(currents[j])

    lookup_currents = np.array(lookup_currents)
    lookup_temperatures = np.array(lookup_temperatures)
    lookup_pressures = np.array(lookup_pressures)
    lookup_voltages = np.array(lookup_voltages)

    # Create lookup table for alkaline electrolyser cell voltage from current
    dict1 = {'temperature in °C' : lookup_temperatures, 'P bar' : lookup_pressures,
             'current in A' : lookup_currents, 'cell voltage in V' : lookup_voltages}
    df1 = pd.DataFrame(dict1)
    df1.to_csv('cell_voltage_lookup_alkaline.csv', index=False)

    # Create lookup table for alkaline electrolyser cell current from power
    lookup_power = lookup_voltages * lookup_currents
    dict2 = {'temperature in °C': lookup_temperatures, 'P bar': lookup_pressures,
             'power in W': lookup_power, 'current in A': lookup_currents}
    df2 = pd.DataFrame(dict2)
    df2.to_csv('cell_current_lookup_alkaline.csv', index=False)


def display_example_curve(temperature: float, pressure: float, compareUI: bool):
    voltages = []
    current_densities = []
    for j in range(len(currents)):
        voltages.append(calculate_cell_voltage(currents[j], temperature, pressure))
        current_densities.append(currents[j] / (anode_surface * 10000))

    fig = plt.figure()

    if compareUI:
        # Compare different electrolyzers
        legend = []
        legend.append('Cell Polarization Curve ' + str(temperature) + '°C at ' + str(pressure) + 'bar')
        ax1 = fig.add_subplot(211)
        ax1.title.set_text('Comparison of different alkaline electrolyzers')
        plt.plot(current_densities, voltages, color='red', linewidth=4)
        plt.ylabel('Cell Voltage in V')
        plt.xlabel('Cell Current Density in A/cm2')
        plt.grid(True)

        u_i_curve_path = '../ui_curves/general_alkaline\\'
        u_i_files = glob.glob(u_i_curve_path + '*.csv')
        styleindex = 0
        for curve in u_i_files:
            styleindex = styleindex + 1
            excelfile = pd.read_csv(curve)
            file_current_densities = excelfile.iloc[:, 0]
            file_cell_voltages = excelfile.iloc[:, 1]
            if styleindex < 11:
                plt.plot(file_current_densities.values, file_cell_voltages.values, linewidth=2)
            else:
                plt.plot(file_current_densities.values, file_cell_voltages.values, linewidth=2, linestyle='dashed')

            legend.append(curve[len(u_i_curve_path):len(curve)-4])

        plt.legend(legend)

        # Compare curves for Stuart HRI Electrolyzer
        legend2 = []
        legend2.append('Cell Polarization Curve ' + str(temperature) + '°C at ' + str(pressure) + 'bar')
        ax2 = fig.add_subplot(212)
        ax2.title.set_text('Comparison of different Stuart HRI Curves at 1 bar')
        plt.plot(current_densities, voltages, color='red', linewidth=4)
        plt.ylabel('Cell Voltage in V')
        plt.xlabel('Cell Current Density in A/cm2')
        plt.grid(True)

        u_i_curve_path = '../ui_curves/stuart_hri\\'
        u_i_files = glob.glob(u_i_curve_path + '*.csv')
        for curve in u_i_files:
            excelfile = pd.read_csv(curve)
            file_current_densities = excelfile.iloc[:, 0]
            file_cell_voltages = excelfile.iloc[:, 1]
            plt.plot(file_current_densities.values, file_cell_voltages.values, linewidth=2)
            legend2.append(curve[len(u_i_curve_path):len(curve) - 4])

        plt.legend(legend2)
    else:
        ax1 = fig.add_subplot(111)
        ax1.title.set_text('Cell Polarisation Curve')
        plt.plot(current_densities, voltages, color='red', linewidth=4)
        plt.ylabel('Cell Voltage in V')
        plt.xlabel('Cell Current Density in A/cm2')
        plt.grid(True)

    plt.show()


def display_example_curve_regression(temperature: float, pressure: float):
    voltages = []
    current_densities = []
    for j in range(len(currents)):
        xdata = np.array([[temperature], [pressure], [currents[j]]])
        voltages.append(cell_voltage_fit_model_function(xdata, fitting_parameters.ac, fitting_parameters.bc, fitting_parameters.cc,
                                                        fitting_parameters.aa, fitting_parameters.ba, fitting_parameters.ca,
                                                        fitting_parameters.r1, fitting_parameters.r2, fitting_parameters.r3,
                                                        fitting_parameters.a1, fitting_parameters.a2))
        current_densities.append(currents[j] / (anode_surface * 10000))

    fig = plt.figure()

    # Compare different electrolyzers
    legend = []
    legend.append('Cell Polarization Curve ' + str(temperature) + '°C at ' + str(pressure) + 'bar')
    ax1 = fig.add_subplot(211)
    ax1.title.set_text('Comparison of different alkaline electrolyzers')
    plt.plot(current_densities, voltages, color='red', linewidth=4)
    plt.ylabel('Cell Voltage in V')
    plt.xlabel('Cell Current Density in A/cm2')
    plt.grid(True)

    u_i_curve_path = '../ui_curves/general_alkaline\\'
    u_i_files = glob.glob(u_i_curve_path + '*.csv')
    styleindex = 0
    for curve in u_i_files:
        styleindex = styleindex + 1
        excelfile = pd.read_csv(curve)
        file_current_densities = excelfile.iloc[:, 0]
        file_cell_voltages = excelfile.iloc[:, 1]
        if styleindex < 11:
            plt.plot(file_current_densities.values, file_cell_voltages.values, linewidth=2)
        else:
            plt.plot(file_current_densities.values, file_cell_voltages.values, linewidth=2, linestyle='dashed')

        legend.append(curve[len(u_i_curve_path):len(curve) - 4])

    plt.legend(legend)

    # Compare curves for Stuart HRI Electrolyzer
    legend = []
    legend.append('Cell Polarization Curve ' + str(temperature) + '°C at ' + str(pressure) + 'bar')
    ax2 = fig.add_subplot(212)
    ax2.title.set_text('Comparison of different Stuart HRI Curves at 1 bar')
    plt.plot(current_densities, voltages, color='red', linewidth=4)
    plt.ylabel('Cell Voltage in V')
    plt.xlabel('Cell Current Density in A/cm2')
    plt.grid(True)

    u_i_curve_path = '../ui_curves/stuart_hri\\'
    u_i_files = glob.glob(u_i_curve_path + '*.csv')
    for curve in u_i_files:
        excelfile = pd.read_csv(curve)
        file_current_densities = excelfile.iloc[:, 0]
        file_cell_voltages = excelfile.iloc[:, 1]
        plt.plot(file_current_densities.values, file_cell_voltages.values, linewidth=2)
        legend.append(curve[len(u_i_curve_path):len(curve) - 4])

    plt.legend(legend)
    plt.show()


def display_activation_overpotential(temperature: float) -> float:
    temperature = temperature + 273.15
    current_densities = []
    v_asinh = []
    v_activation_lookup = []
    currents_activation = np.linspace(0, 100, 300)
    for j in range(len(currents_activation)):
        if currents_activation[j] > 0:
            current_densities.append(currents_activation[j] / (membrane_surface_area * 10000))
            v_act_a = calculate_activation_overvoltage_anode(currents_activation[j], temperature)
            v_act_c = calculate_activation_overvoltage_cathode(currents_activation[j], temperature)
            v_act_total = v_act_a + v_act_c

            v_act_total_lookup = calculate_activation_overvoltage_lookup(currents_activation[j], temperature)
            v_asinh.append(v_act_total)
            v_activation_lookup.append(v_act_total_lookup)
        else:
            current_densities.append(currents_activation[j] / (membrane_surface_area * 10000))
            v_asinh.append(0)
            v_activation_lookup.append(0)

    # Deviation Analysis
    absolute_error = np.array(v_asinh) - np.array(v_activation_lookup)
    mean_absolute_error = np.mean(abs(absolute_error))
    rms_error = np.sqrt(sum((absolute_error) ** 2) / len(absolute_error))

    # Plotting
    plt.subplot(111)
    plt.ylabel('Activation Voltage V')
    plt.xlabel('Cell Current Density in A/cm2')
    plt.grid(True)
    plt.title('Activation Overvoltage ' + str(temperature - 273.15) + '°C')
    plt.plot(current_densities, v_asinh)
    plt.plot(current_densities, v_activation_lookup)
    plt.legend(["v_act_asinh", "v_act_lookup (butler volmer)"])
    plt.show()

    return rms_error


# Three fitting parameters each for quadratic description of exchange current densities as f(T) (As aforementioned paper by Henao et al)
def cell_voltage_fit_model_function(x, ac, bc, cc, aa, ba, ca, r1, r2, r3, a1, a2):
    temperature = x[0, :]
    pressure = x[1, :]
    current = x[2, :]

    v_cell = 0

    # Convert temperature from °C to K
    temperature = (temperature + 273.15)

    # Get theoretical cell voltage
    v_theoretical = calculate_theoretical_voltage(temperature, pressure)
    v_cell += v_theoretical

    # Calculate bubble rate coverage of electrodes (assumed to be identical for both electrodes)
    bubble_rate_coverage = calculate_bubble_rate_coverage(current, temperature)
    # Calculate nominal electrode surfaces
    anode_surface_nom = anode_surface * (1 - bubble_rate_coverage)
    current_density_anode = current / (anode_surface_nom * 10000.0)  # A/cm2
    cathode_surface_nom = cathode_surface * (1 - bubble_rate_coverage)
    current_density_cathode = current / (cathode_surface_nom * 10000.0)
    # Calculate transfer coefficient for anode and cathode for HRI Electrolyser
    alpha_a = 0.0675 + 0.00095 * temperature*a1
    alpha_c = 0.1175 + 0.00095 * temperature*a2
    # Calculate tafel slopes for anode and cathode
    factor_anode = (__IDEAL_GAS_CONST * temperature) / (
                2 * __FARADAY_CONST * alpha_a)
    factor_cathode = (__IDEAL_GAS_CONST * temperature) / (
            2 * __FARADAY_CONST * alpha_c)
    # Calculate exchange current densities for anode and cathode for HRI Electrolyser (Ni Electrodes)
    exchange_current_density_anode = aa + ba * temperature + ca * temperature ** 2  # mA/cm2
    exchange_current_density_cathode = ac + bc * temperature + cc * temperature ** 2  # mA/cm2
    # Calculate activation overvoltages of the electrodes:
    v_activation_anode = factor_anode * np.arcsinh(current_density_anode / (2 * exchange_current_density_anode / 1000.0))
    v_activation_cathode = factor_cathode * np.arcsinh(current_density_cathode / (2 * exchange_current_density_cathode / 1000.0))
    v_activation_total = v_activation_anode + v_activation_cathode
    v_cell += v_activation_total
    # Calculate ohmic resistances
    # Electrolyte
    v_electrolyte_resistance = calculate_electrolyte_overvoltage(current, temperature)
    v_cell += v_electrolyte_resistance * r1
    # Electrodes
    v_electrode_resistances = calculate_electrode_overvoltage(current, temperature)
    v_cell += v_electrode_resistances * r2
    # Membrane
    v_membrane_resistance = calculate_membrane_overvoltage(current, temperature)
    v_cell += v_membrane_resistance * r3

    return v_cell


def get_fitting_data():
    u_i_curve_path = '../ui_curves/stuart_hri\\'
    u_i_files = glob.glob(u_i_curve_path + '*.csv')
    currents_fit = []
    cell_voltages_fit = []
    pressures_fit = []
    temperatures_fit = []
    for curve in u_i_files:
        excelfile = pd.read_csv(curve)
        file_current_densities = excelfile.iloc[:, 0]
        file_cell_voltages = excelfile.iloc[:, 1]
        file_temperatures = excelfile.iloc[:, 2]
        file_pressures = excelfile.iloc[:, 3]
        currents_fit = np.concatenate([currents_fit, file_current_densities * cathode_surface * 10000])
        cell_voltages_fit = np.concatenate([cell_voltages_fit, file_cell_voltages])
        temperatures_fit = np.concatenate([temperatures_fit, file_temperatures])
        pressures_fit = np.concatenate([pressures_fit, file_pressures])

    # Extra Curves for other alkaline electrolyzers
    # u_i_curve_path = 'UICurves\\GeneralAlkaline\\'
    # u_i_files = glob.glob(u_i_curve_path + '*.csv')
    # for curve in u_i_files:
    #     excelfile = pd.read_csv(curve)
    #     file_current_densities = excelfile.iloc[:, 0]
    #     file_cell_voltages = excelfile.iloc[:, 1]
    #     file_temperatures = excelfile.iloc[:, 2]
    #     file_pressures = excelfile.iloc[:, 3]
    #     currents_fit = np.concatenate([currents_fit, file_current_densities * cathode_surface * 10000])
    #     cell_voltages_fit = np.concatenate([cell_voltages_fit, file_cell_voltages])
    #     temperatures_fit = np.concatenate([temperatures_fit, file_temperatures])
    #     pressures_fit = np.concatenate([pressures_fit, file_pressures])

    xdata = np.array([temperatures_fit, pressures_fit, currents_fit])
    ydata = np.array(cell_voltages_fit)

    return xdata, ydata

def polarisation_deviations():
    xdata_ref, voltages_ref = get_fitting_data()
    temperatures_ref = xdata_ref[0]
    pressures_ref = xdata_ref[1]
    currents_ref = xdata_ref[2]

    voltages_calculated = []
    for i in range(len(temperatures_ref)):
        xdata = np.array([[temperatures_ref[i]], [pressures_ref[i]], [currents_ref[i]]])
        voltages_calculated.append(cell_voltage_fit_model_function(xdata, fitting_parameters.ac, fitting_parameters.bc,
                                                        fitting_parameters.cc,
                                                        fitting_parameters.aa, fitting_parameters.ba,
                                                        fitting_parameters.ca,
                                                        fitting_parameters.r1, fitting_parameters.r2,
                                                        fitting_parameters.r3,
                                                        fitting_parameters.a1, fitting_parameters.a2))
    voltages_calculated = np.array(voltages_calculated)
    voltages_calculated = voltages_calculated.reshape(len(voltages_calculated),)
    #voltages_ref = voltages_ref.reshape(len(voltages_ref), 1)
    #currents_ref = currents_ref.reshape(len(currents_ref), 1)
    voltage_error_absolute = abs(voltages_calculated - voltages_ref)
    mean_abs_err = np.mean(voltage_error_absolute)
    max_abs_err = max(abs(voltage_error_absolute))
    RMS = np.sqrt(sum((voltage_error_absolute) ** 2) / len(voltage_error_absolute))
    print(RMS)

    fig = plt.figure()
    legend = []
    ax1 = fig.add_subplot(111)
    for temperature in np.unique(temperatures_ref):
        currents_selected = currents_ref[np.where(temperatures_ref == temperature)]
        voltage_error_selected = voltage_error_absolute[np.where(temperatures_ref == temperature)]
        plt.plot(currents_selected, voltage_error_selected, linewidth=2)
        # Export Error Data in Excel
        # dict1 = {'Current in A': currents_selected, 'Absolute Error in V': voltage_error_selected}
        # df1 = pd.DataFrame(dict1)
        # df1.to_csv(('Absolute_Error_' + str(temperature) + '.csv'), index=False)

    plt.ylabel('Absolute Voltage Error in V')
    plt.xlabel('Cell Current in A')
    plt.grid(True)
    plt.show()


# COMMAND SECTION

# xdata, ydata = get_fitting_data()
# lb = - np.inf
# ub = np.inf
# EPS = np.finfo(float).eps
# lower_bounds = (lb, lb, lb, lb, lb, lb, lb, lb, lb, lb, lb)
# upper_bounds = (ub, ub, ub, ub, ub, ub, ub, ub, ub, ub, ub)
# popt, pcov = curve_fit(cell_voltage_fit_model_function, xdata, ydata, absolute_sigma=False, p0=[1, 0, 0, 1, 0, 0, 1, 1, 1, 1, 1], maxfev=1000000,
#                        bounds=(lower_bounds,upper_bounds))
# dict = {'parameters': popt}
# df = pd.DataFrame(dict)
# df.to_csv('parameters_alkaline_overvoltage_fit.csv', index=False)

# display_example_curve_regression(53, 1)

# generate_polarisation_lookup()

# display_example_curve(23, 1, compareUI = True)

# rms_error = []
# for i in range(len(temperatures)):
#     rms_error.append(display_activation_overpotential(temperatures[i]))
#
#
# avg_rms_error = sum(rms_error) / len(rms_error)
# print(avg_rms_error)

polarisation_deviations()




import pandas as pd

from simses.commons.config.data.electrolyzer import ElectrolyzerDataConfig


class ParametersPemElectrolyzerMultiDimAnalytic:

    def __init__(self, electrolyzer_data_config: ElectrolyzerDataConfig):
        super().__init__()
        file = electrolyzer_data_config.pem_electrolyzer_multi_dim_analytic_para_file
        correction_parameters = pd.read_csv(file)
        self.__corr_parameters = correction_parameters.iloc[:, 0]
        # parameters of quadratic regression of activation powerdensity
        self.__p10 = -0.007005
        self.__p20 = 0.01346
        self.__p11 = -0.0001083

    @property
    def q1(self) -> float:
        return self.__corr_parameters[0]

    @property
    def q3(self) -> float:
        return self.__corr_parameters[1]

    @property
    def q4(self) -> float:
        return self.__corr_parameters[2]

    @property
    def q5(self) -> float:
        return self.__corr_parameters[3]

    @property
    def q6(self) -> float:
        return self.__corr_parameters[4]

    @property
    def q7(self) -> float:
        return self.__corr_parameters[5]

    @property
    def q9(self) -> float:
        return self.__corr_parameters[6]

    @property
    def q10(self) -> float:
        return self.__corr_parameters[7]

    @property
    def q11(self) -> float:
        return self.__corr_parameters[8]

    @property
    def q12(self) -> float:
        return self.__corr_parameters[9]

    @property
    def q13(self) -> float:
        return self.__corr_parameters[10]

    @property
    def q14(self) -> float:
        return self.__corr_parameters[11]

    @property
    def q15(self) -> float:
        return self.__corr_parameters[12]

    @property
    def q17(self) -> float:
        return self.__corr_parameters[13]

    @property
    def p10(self) -> float:
        return self.__p10

    @property
    def p20(self) -> float:
        return self.__p20

    @property
    def p11(self) -> float:
        return self.__p11


class EndOfLifeError(Exception):
    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return '{0}'.format(self.message)
        else:
            return 'End of life has been reached'


class EndOfFileError(Exception):
    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return '{0}'.format(self.message)
        else:
            return 'End of file has been reached'


class InfiniteNumberError(Exception):
    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return '{0}'.format(self.message)
        else:
            return 'Number is infinite'


class NanNumberError(Exception):
    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return '{0}'.format(self.message)
        else:
            return 'Number is NaN'


class ComplexNumberError(Exception):
    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return '{0}'.format(self.message)
        else:
            return 'Number is complex'



from simses.logic.thermal_management.thermal_management import ThermalManagement
from simses.system.auxiliary.heating_ventilation_air_conditioning.fan import Fan
from simses.system.thermal.model.zero_d_system_thermal_model import ZeroDSystemThermalModel
from numpy import sign
import pandas as pd
import os
import time


class PIDController(ThermalManagement):

    TEMPERATURE_DEAD_BAND: float = 2  # K

    def __init__(self, set_point: float, hvac_max_thermal_power: float):
        super().__init__()
        self.__export_running_data = False
        self.__set_temperature = set_point  # K
        self.__hvac_rated_thermal_power = hvac_max_thermal_power  # W
        self.__kp_coefficient = 8000
        self.__ki_coefficient = 200
        self.__kd_coefficient = 1800
        self.__thermal_power_scaling_factor = 1
        self.__temperature_difference_memory = [0]
        self.__i_temperature_difference = 0

        # Export running data
        if self.__export_running_data:
            self.__file_name = os.getcwd() + '\\controller_running_data\\Results' + str(time.time()) + '.csv'
            df = pd.DataFrame(['Ti', 'Ta', 'Total Thermal Power', 'HVAC Thermal Power', 'Fan Power', 'Required thermal power', 'P_Difference', 'I_Difference', 'D_Difference', 'HVAC Limit']).T
            df.to_csv(self.__file_name, header=None, index=None)

    def compute(self, internal_air_temperature: float, ambient_air_temperature: float, air_mass: float,
                air_density: float, time_step: float, fan: Fan) -> list:
        p_temperature_difference = float(internal_air_temperature - self.__set_temperature)
        self.__i_temperature_difference += p_temperature_difference
        self.__temperature_difference_memory.append(p_temperature_difference)
        d_temperature_difference = (self.__temperature_difference_memory[-1] - self.__temperature_difference_memory[-2])

        required_thermal_power = self.__thermal_power_scaling_factor * \
                                 (self.__kp_coefficient * p_temperature_difference +
                                  self.__ki_coefficient * self.__i_temperature_difference +
                                  self.__kd_coefficient * d_temperature_difference)  # W

        if sign(required_thermal_power) == sign(p_temperature_difference) and abs(p_temperature_difference) > self.TEMPERATURE_DEAD_BAND:
            # -ve thermal power: heating, +ve thermal power: cooling
            # Filter with deadband to prevent overcooling / overheating

            rated_supply_temperature = float(self.__set_temperature - sign(required_thermal_power) * 10)  # K
            delta_temperature_internal_ambient = float(internal_air_temperature - ambient_air_temperature)  # K
            rated_fresh_air_thermal_power = float(fan.rated_airflow * air_density * delta_temperature_internal_ambient * ZeroDSystemThermalModel.AIR_SPECIFIC_HEAT)  # W
            max_physical_fresh_air_thermal_power = float(air_mass * delta_temperature_internal_ambient * ZeroDSystemThermalModel.AIR_SPECIFIC_HEAT / time_step)  # W

            if sign(required_thermal_power) == sign(delta_temperature_internal_ambient):
                # checks if the fresh air thermal power is usable
                # -ve delta_temperature_internal_ambient: outside hotter, +ve: outside cooler
                fresh_air_thermal_power = min(min(abs(max_physical_fresh_air_thermal_power), abs(rated_fresh_air_thermal_power)), abs(required_thermal_power)) * float(sign(required_thermal_power))  # W
                if fresh_air_thermal_power / required_thermal_power >= 0.999:
                    # first case: the requirement can be met only with fresh air
                    hvac_thermal_power = 0.0  # W
                else:
                    # second case: fresh air thermal power is not enough, use HVAC
                    if abs(rated_supply_temperature) - abs(ambient_air_temperature) > 0.1:
                        hvac_thermal_power = float(min(abs(self.__hvac_rated_thermal_power),
                                                       abs(required_thermal_power) - abs(fresh_air_thermal_power)) * sign(required_thermal_power))  # W
                    else:
                        hvac_thermal_power = 0.0

                airflow_ratio = (fresh_air_thermal_power / rated_fresh_air_thermal_power)  # m3/s
                if airflow_ratio <= 1:
                    airflow = fan.rated_airflow * (fresh_air_thermal_power / rated_fresh_air_thermal_power)  # m3/s
                else:
                    airflow = fan.rated_airflow
            else:
                # third case: fresh air can not be used. HVAC only.
                fresh_air_thermal_power = 0.0  # W
                if abs(rated_supply_temperature) - abs(ambient_air_temperature) > 0.1:
                    hvac_thermal_power = float(min(abs(self.__hvac_rated_thermal_power), abs(required_thermal_power)) * sign(required_thermal_power))  # W
                else:
                    hvac_thermal_power = 0.0
                airflow = 0.0  # m3/s
        else:
            hvac_thermal_power = 0.0  # W
            fresh_air_thermal_power = 0.0  # W
            airflow = 0.0  # m3/s
        total_thermal_power = float(hvac_thermal_power + fresh_air_thermal_power)  # W
        fan.run(airflow)

        # Export running data
        if self.__export_running_data:
            data_list = [float(internal_air_temperature), ambient_air_temperature, total_thermal_power, hvac_thermal_power,
                        fan.electricity_consumption, required_thermal_power, p_temperature_difference,
                        self.__i_temperature_difference, d_temperature_difference, self.__hvac_rated_thermal_power]
            df = pd.DataFrame(data_list).T
            df.to_csv(self.__file_name, mode='a', header=None, index=None)

        return [float(total_thermal_power), float(hvac_thermal_power)]




from xml.etree.ElementTree import Element

import numpy

from simses.commons.utils.xml_reader import XmlReader
from simses.technology.lithium_ion.cell.electric.properties import ElectricalCellProperties


class IseaCellReader:
    # OLD i3Cell-Data:
    # __EQUIVALENT_CIRCUIT_MODEL: str = 'CustomDefinitions'

    # Kokam Cell Start:
    __GENERAL_INFO: str = 'GeneralInformation'
    __ELECTRICAL_PARAMETER: str = 'ElectricalParameters'
    __NOMINAL_CAPACITY: str = 'NominalCapacity'
    __VALUE: str = "Value"
    __NOMINAL_VOLTAGE: str = 'NominalCellVoltage'
    __MAX_CHARGE_CURRENT: str = 'MaxChargeCurrentCont'
    __MAX_DISCHARGE_CURRENT: str = 'MaxDischargeCurrentCont'
    __MAX_VOLTAGE: str = 'EndOfChargeVoltage'
    __MIN_VOLTAGE: str = 'EndOfDischargeVoltage'

    __MODEL_LEVEL: str = 'ModelLevel'
    __ELECTRICAL_MODEL: str = 'ElectricalModel'
    __ISEA_R2RC_DATA: str = 'ISEA_R_OCV'
    __CONFIGURATION: str = 'Configuration'
    __CUSTOM_DEFINITION: str = 'CustomDefinitions'
    # Kokam Cell End

    __OPEN_CIRCUIT_VOLTAGE: str = 'MyOCV'
    __OPEN_CIRCUIT_VOLTAGE_OBJECT: str = 'Object'
    __OCV_DATA: str = 'LookupData'
    __OCV_SOC_DATA: str = 'MeasurementPointsRow'
    __OCV_TEMPERATURE_DATA: str = 'MeasurementPointsColumn'

    __INTERNAL_RESISTANCE: str = 'MyOhmicResistanceser'
    __INTERNAL_RESISTANCE_OBJECT: str = 'Object'
    __INTERNAL_RESISTANCE_DATA: str = 'LookupData'
    __INTERNAL_RESISTANCE_SOC_DATA: str = 'MeasurementPointsRow'
    __INTERNAL_RESISTANCE_TEMPERATURE_DATA: str = 'MeasurementPointsColumn'

    def __init__(self, path: str):
        self.__data_cell: XmlReader = XmlReader(path)

    def __get_equivalent_circuit_model(self) -> Element:
        # Kokam Cell Structure:
        model_level: Element = self.__data_cell.get_element(self.__MODEL_LEVEL)
        electrical_model: Element = self.__data_cell.get_element(self.__ELECTRICAL_MODEL, model_level)
        isea_data: Element = self.__data_cell.get_element(self.__ISEA_R2RC_DATA, electrical_model)
        configuration_data: Element = self.__data_cell.get_element(self.__CONFIGURATION, isea_data)
        custom_def: Element = self.__data_cell.get_element(self.__CUSTOM_DEFINITION, configuration_data)
        return custom_def

    def __get_open_circuit_voltage(self) -> Element:
        ecm: Element = self.__get_equivalent_circuit_model()
        ocv: Element = self.__data_cell.get_element(self.__OPEN_CIRCUIT_VOLTAGE, ecm)
        return self.__data_cell.get_element(self.__OPEN_CIRCUIT_VOLTAGE_OBJECT, ocv)

    def __get_internal_resistance(self) -> Element:
        ecm: Element = self.__get_equivalent_circuit_model()
        rint: Element = self.__data_cell.get_element(self.__INTERNAL_RESISTANCE, ecm)
        return self.__data_cell.get_element(self.__INTERNAL_RESISTANCE_OBJECT, rint)

    def get_cell_properties(self):
        general_info: Element = self.__data_cell.get_element(self.__GENERAL_INFO)
        electrical_param: Element = self.__data_cell.get_element(self.__ELECTRICAL_PARAMETER, general_info)

        nom_capacity: Element = self.__data_cell.get_element(self.__NOMINAL_CAPACITY, electrical_param)
        nom_capacity_value: Element = self.__data_cell.get_element(self.__VALUE, nom_capacity)
        cell_capacity = self.__data_cell.parse(nom_capacity_value).item()  # Ah

        nom_voltage: Element = self.__data_cell.get_element(self.__NOMINAL_VOLTAGE, electrical_param)
        nom_voltage_value: Element = self.__data_cell.get_element(self.__VALUE, nom_voltage)
        cell_voltage = self.__data_cell.parse(nom_voltage_value).item()  # V

        max_charge: Element = self.__data_cell.get_element(self.__MAX_CHARGE_CURRENT, electrical_param)
        max_charge_value: Element = self.__data_cell.get_element(self.__VALUE, max_charge)
        try:
            max_charge_current = self.__data_cell.parse(max_charge_value).item()  # A
            max_charge_rate = max_charge_current / cell_capacity  # 1/h
        except ValueError: # if no value available
            max_charge_rate = 0.5

        max_discharge: Element = self.__data_cell.get_element(self.__MAX_DISCHARGE_CURRENT, electrical_param)
        max_discharge_value: Element = self.__data_cell.get_element(self.__VALUE, max_discharge)
        try:
            max_discharge_current = self.__data_cell.parse(max_discharge_value).item()  # A
            max_discharge_rate = max_discharge_current / cell_capacity  # 1/h
        except ValueError: # if no value available
            max_discharge_rate = 0.5


        max_voltage_data: Element = self.__data_cell.get_element(self.__MAX_VOLTAGE, electrical_param)
        max_voltage_value: Element = self.__data_cell.get_element(self.__VALUE, max_voltage_data)
        max_voltage = self.__data_cell.parse(max_voltage_value).item()  # V

        min_voltage_data: Element = self.__data_cell.get_element(self.__MIN_VOLTAGE, electrical_param)
        min_voltage_value: Element = self.__data_cell.get_element(self.__VALUE, min_voltage_data)
        min_voltage = self.__data_cell.parse(min_voltage_value).item()  # V

        # cell_voltage = 7  # V
        # cell_capacity = 2.5  # Ah
        # max_voltage: float = 4.0  # V
        # min_voltage: float = 3.0  # V
        # max_charge_rate: float = 2.0  # 1/h
        # max_discharge_rate: float = 2.0  # 1/h
        self_discharge_rate: float = 0.0  # X.X%-soc per day, e.g., 0.015 for 1.5% SOC loss per day
        coulomb_efficiency: float = 1.0  # p.u.

        isea_cell = ElectricalCellProperties(cell_voltage, cell_capacity,
                                             min_voltage, max_voltage,
                                             max_charge_rate, max_discharge_rate,
                                             self_discharge_rate, coulomb_efficiency)

        return isea_cell

    def get_open_cicuit_voltage(self) -> numpy.ndarray:
        ocv: Element = self.__get_open_circuit_voltage()
        ocv_data: Element = self.__data_cell.get_element(self.__OCV_DATA, ocv)
        return self.__data_cell.parse(ocv_data)

    def get_open_cicuit_voltage_soc(self) -> numpy.ndarray:
        ocv: Element = self.__get_open_circuit_voltage()
        soc_data: Element = self.__data_cell.get_element(self.__OCV_SOC_DATA, ocv)
        return self.__data_cell.parse(soc_data)

    def get_open_cicuit_voltage_temperature(self) -> numpy.ndarray:
        ocv: Element = self.__get_open_circuit_voltage()
        temperature_data: Element = self.__data_cell.get_element(self.__OCV_TEMPERATURE_DATA, ocv)
        return self.__data_cell.parse(temperature_data)

    def get_internal_resistance(self) -> numpy.ndarray:
        rint: Element = self.__get_internal_resistance()
        ocv_data: Element = self.__data_cell.get_element(self.__INTERNAL_RESISTANCE_DATA, rint)
        return self.__data_cell.parse(ocv_data)

    def get_internal_resistance_soc(self) -> numpy.ndarray:
        rint: Element = self.__get_internal_resistance()
        soc_data: Element = self.__data_cell.get_element(self.__INTERNAL_RESISTANCE_SOC_DATA, rint)
        return self.__data_cell.parse(soc_data)

    def get_internal_resistance_temperature(self) -> numpy.ndarray:
        rint: Element = self.__get_internal_resistance()
        temperature_data: Element = self.__data_cell.get_element(self.__INTERNAL_RESISTANCE_TEMPERATURE_DATA, rint)
        return self.__data_cell.parse(temperature_data)


if __name__ == '__main__':
    path: str = '../../data/lithium_ion/isea/i3Cell.xml'
    cell_data: IseaCellReader = IseaCellReader(path)
    ocv_values: numpy.ndarray = cell_data.get_open_cicuit_voltage()
    soc_values: numpy.ndarray = cell_data.get_open_cicuit_voltage_soc()
    print(type(ocv_values), ocv_values)
    print(type(soc_values), soc_values)
